#  Copyright (c) ZenML GmbH 2021. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Base Step for ZenML."""

import copy
import hashlib
import inspect
from abc import abstractmethod
from collections import defaultdict
from contextlib import contextmanager
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Generator,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)
from uuid import UUID

from pydantic import BaseModel, ConfigDict, ValidationError

from zenml.client_lazy_loader import ClientLazyLoader
from zenml.config.cache_policy import CachePolicyOrString
from zenml.config.retry_config import StepRetryConfig
from zenml.config.source import Source
from zenml.constants import (
    CODE_HASH_PARAMETER_NAME,
    ENV_ZENML_RUN_SINGLE_STEPS_WITHOUT_STACK,
    handle_bool_env_var,
)
from zenml.enums import GroupType, StepRuntime, StepType
from zenml.exceptions import SourceValidationException, StepInterfaceError
from zenml.logger import get_logger
from zenml.materializers.base_materializer import BaseMaterializer
from zenml.materializers.materializer_registry import materializer_registry
from zenml.steps.entrypoint_function_utils import (
    StepArtifact,
    validate_entrypoint_function,
)
from zenml.steps.utils import (
    resolve_type_annotation,
    run_as_single_step_pipeline,
)
from zenml.utils import (
    dict_utils,
    materializer_utils,
    notebook_utils,
    pydantic_utils,
    settings_utils,
    source_code_utils,
    source_utils,
)

if TYPE_CHECKING:
    from zenml.artifacts.external_artifact import ExternalArtifact
    from zenml.artifacts.external_artifact_config import (
        ExternalArtifactConfiguration,
    )
    from zenml.config.base_settings import SettingsOrDict
    from zenml.config.step_configurations import (
        GroupInfo,
        PartialArtifactConfiguration,
        PartialStepConfiguration,
        StepConfiguration,
        StepConfigurationUpdate,
    )
    from zenml.model.lazy_load import ModelVersionDataLazyLoader
    from zenml.model.model import Model
    from zenml.models import ArtifactVersionResponse
    from zenml.types import HookSpecification

    MaterializerClassOrSource = Union[str, Source, Type["BaseMaterializer"]]
    OutputMaterializersSpecification = Union[
        "MaterializerClassOrSource",
        Sequence["MaterializerClassOrSource"],
        Mapping[str, "MaterializerClassOrSource"],
        Mapping[str, Sequence["MaterializerClassOrSource"]],
    ]

    from zenml.execution.pipeline.dynamic.outputs import (
        AnyStepFuture,
        MapResultsFuture,
        StepFuture,
    )
    from zenml.pipelines.compilation_context import (
        PipelineCompilationContext,
    )


logger = get_logger(__name__)

T = TypeVar("T", bound="BaseStep")


class BaseStep:
    """Abstract base class for all ZenML steps."""

    def __init__(
        self,
        name: Optional[str] = None,
        step_type: Optional[StepType] = None,
        enable_cache: Optional[bool] = None,
        enable_artifact_metadata: Optional[bool] = None,
        enable_artifact_visualization: Optional[bool] = None,
        enable_step_logs: Optional[bool] = None,
        experiment_tracker: Optional[Union[bool, str]] = None,
        step_operator: Optional[Union[bool, str]] = None,
        parameters: Optional[Dict[str, Any]] = None,
        output_materializers: Optional[
            "OutputMaterializersSpecification"
        ] = None,
        environment: Optional[Dict[str, Any]] = None,
        secrets: Optional[List[Union[str, UUID]]] = None,
        settings: Optional[Mapping[str, "SettingsOrDict"]] = None,
        extra: Optional[Dict[str, Any]] = None,
        on_failure: Optional["HookSpecification"] = None,
        on_success: Optional["HookSpecification"] = None,
        model: Optional["Model"] = None,
        retry: Optional[StepRetryConfig] = None,
        substitutions: Optional[Dict[str, str]] = None,
        cache_policy: Optional[CachePolicyOrString] = None,
        runtime: Optional[StepRuntime] = None,
        heartbeat_healthy_threshold: Optional[int] = None,
        group: Optional[Union["GroupInfo", str]] = None,
    ) -> None:
        """Initializes a step.

        Args:
            name: The name of the step.
            step_type: The type of the step.
            enable_cache: If caching should be enabled for this step.
            enable_artifact_metadata: If artifact metadata should be enabled
                for this step.
            enable_artifact_visualization: If artifact visualization should be
                enabled for this step.
            enable_step_logs: Enable step logs for this step.
            experiment_tracker: The experiment tracker to use for this step.
            step_operator: The step operator to use for this step.
            parameters: Function parameters for this step
            output_materializers: Output materializers for this step. If
                given as a dict, the keys must be a subset of the output names
                of this step. If a single value (type or string) is given, the
                materializer will be used for all outputs.
            environment: Environment variables to set when running this step.
            secrets: Secrets to set as environment variables when running this
                step.
            settings: Settings for this step.
            extra: Extra configurations for this step.
            on_failure: Callback function in event of failure of the step. Can
                be a function with a single argument of type `BaseException`, or
                a source path to such a function (e.g. `module.my_function`).
            on_success: Callback function in event of success of the step. Can
                be a function with no arguments, or a source path to such a
                function (e.g. `module.my_function`).
            model: configuration of the model version in the Model Control Plane.
            retry: Configuration for retrying the step in case of failure.
            substitutions: Extra substitutions for model and artifact name
                placeholders.
            cache_policy: Cache policy for this step.
            runtime: The step runtime. If not configured, the step will
                run inline unless a step operator or docker/resource settings
                are configured. This is only applicable for dynamic
                pipelines.
            heartbeat_healthy_threshold: The amount of time (in minutes) that a
                running step has not received heartbeat and is considered healthy.
                By default, set to 30 minutes.",
            group: The group information for this step.
        """
        from zenml.config.step_configurations import PartialStepConfiguration

        self.entrypoint_definition = validate_entrypoint_function(
            self.entrypoint,
            reserved_arguments=["after", "id"],
        )

        self._static_id = id(self)
        name = name or self.__class__.__name__

        logger.debug(
            "Step `%s`: Caching %s.",
            name,
            "enabled" if enable_cache is not False else "disabled",
        )
        logger.debug(
            "Step `%s`: Artifact metadata %s.",
            name,
            "enabled" if enable_artifact_metadata is not False else "disabled",
        )
        logger.debug(
            "Step `%s`: Artifact visualization %s.",
            name,
            "enabled"
            if enable_artifact_visualization is not False
            else "disabled",
        )
        logger.debug(
            "Step `%s`: logs %s.",
            name,
            "enabled" if enable_step_logs is not False else "disabled",
        )
        if model is not None:
            logger.debug(
                "Step `%s`: Is in Model context %s.",
                name,
                {
                    "model": model.name,
                    "version": model.version,
                },
            )

        self._configuration = PartialStepConfiguration(
            name=name, step_type=step_type
        )
        self._dynamic_configuration: Optional["StepConfigurationUpdate"] = None
        self._capture_dynamic_configuration = True

        self.configure(
            enable_cache=enable_cache,
            enable_artifact_metadata=enable_artifact_metadata,
            enable_artifact_visualization=enable_artifact_visualization,
            enable_step_logs=enable_step_logs,
            experiment_tracker=experiment_tracker,
            step_operator=step_operator,
            output_materializers=output_materializers,
            parameters=parameters,
            environment=environment,
            secrets=secrets,
            settings=settings,
            extra=extra,
            on_failure=on_failure,
            on_success=on_success,
            model=model,
            retry=retry,
            substitutions=substitutions,
            cache_policy=cache_policy,
            runtime=runtime,
            heartbeat_healthy_threshold=heartbeat_healthy_threshold,
            group=group,
        )

        notebook_utils.try_to_save_notebook_cell_code(self.source_object)

    @abstractmethod
    def entrypoint(self, *args: Any, **kwargs: Any) -> Any:
        """Abstract method for core step logic.

        Args:
            *args: Positional arguments passed to the step.
            **kwargs: Keyword arguments passed to the step.

        Returns:
            The output of the step.
        """

    @classmethod
    def load_from_source(cls, source: Union[Source, str]) -> "BaseStep":
        """Loads a step from source.

        Args:
            source: The path to the step source.

        Returns:
            The loaded step.

        Raises:
            ValueError: If the source is not a valid step source.
        """
        obj = source_utils.load(source)

        if isinstance(obj, BaseStep):
            return obj
        elif isinstance(obj, type) and issubclass(obj, BaseStep):
            return obj()
        else:
            raise ValueError("Invalid step source.")

    def resolve(self) -> Source:
        """Resolves the step.

        Returns:
            The step source.
        """
        return source_utils.resolve(self.__class__)

    @property
    def source_object(self) -> Any:
        """The source object of this step.

        Returns:
            The source object of this step.
        """
        return self.__class__

    @property
    def source_code(self) -> str:
        """The source code of this step.

        Returns:
            The source code of this step.
        """
        return inspect.getsource(self.source_object)

    @property
    def source_code_cache_value(self) -> str:
        """The source code cache value of this step.

        Returns:
            The source code cache value of this step.
        """
        return self.source_code

    @property
    def docstring(self) -> Optional[str]:
        """The docstring of this step.

        Returns:
            The docstring of this step.
        """
        return self.__doc__

    @property
    def caching_parameters(self) -> Dict[str, Any]:
        """Caching parameters for this step.

        Returns:
            A dictionary containing the caching parameters
        """
        parameters = {
            CODE_HASH_PARAMETER_NAME: hashlib.sha256(
                self.source_code_cache_value.encode("utf-8")
            ).hexdigest()
        }
        for name, output in self.configuration.outputs.items():
            if output.materializer_source:
                key = f"{name}_materializer_source"
                hash_ = hashlib.md5()  # nosec

                for source in output.materializer_source:
                    materializer_class = source_utils.load(source)
                    code_hash = source_code_utils.get_hashed_source_code(
                        materializer_class
                    )
                    hash_.update(code_hash.encode())

                parameters[key] = hash_.hexdigest()

        return parameters

    def _parse_call_args(
        self, *args: Any, **kwargs: Any
    ) -> Tuple[
        Dict[str, List["StepArtifact"]],
        Dict[str, Union["ExternalArtifact", "ArtifactVersionResponse"]],
        Dict[str, "ModelVersionDataLazyLoader"],
        Dict[str, "ClientLazyLoader"],
        Dict[str, Any],
        Dict[str, Any],
    ]:
        """Parses the call args for the step entrypoint.

        Args:
            *args: Entrypoint function arguments.
            **kwargs: Entrypoint function keyword arguments.

        Raises:
            StepInterfaceError: If invalid function arguments were passed.

        Returns:
            The artifacts, external artifacts, model version artifacts/metadata and parameters for the step.
        """
        from zenml.artifacts.external_artifact import ExternalArtifact
        from zenml.metadata.lazy_load import LazyRunMetadataResponse
        from zenml.model.lazy_load import ModelVersionDataLazyLoader
        from zenml.models.v2.core.artifact_version import (
            ArtifactVersionResponse,
            LazyArtifactVersionResponse,
        )

        signature = inspect.signature(self.entrypoint, follow_wrapped=True)

        try:
            bound_args = signature.bind_partial(*args, **kwargs)
        except TypeError as e:
            raise StepInterfaceError(
                f"Wrong arguments when calling step '{self.name}': {e}"
            ) from e

        artifacts = {}
        external_artifacts: Dict[
            str, Union["ExternalArtifact", "ArtifactVersionResponse"]
        ] = {}
        model_artifacts_or_metadata = {}
        client_lazy_loaders = {}
        parameters = {}
        default_parameters = {}

        for key, value in bound_args.arguments.items():
            self.entrypoint_definition.validate_input(key=key, value=value)

            if isinstance(value, StepArtifact):
                artifacts[key] = [value]
                if key in self.configuration.parameters:
                    logger.warning(
                        "Got duplicate value for step input %s, using value "
                        "provided as artifact.",
                        key,
                    )
            elif (
                isinstance(value, list)
                and value
                and all(isinstance(item, StepArtifact) for item in value)
            ):
                artifacts[key] = value
                if key in self.configuration.parameters:
                    logger.warning(
                        "Got duplicate value for step input %s, using value "
                        "provided as artifact.",
                        key,
                    )
            elif isinstance(value, ExternalArtifact):
                external_artifacts[key] = value
                if not value.id:
                    # If the external artifact references a fixed artifact by
                    # ID, caching behaves as expected.
                    logger.warning(
                        "Using an external artifact as step input currently "
                        "invalidates caching for the step and all downstream "
                        "steps. Future releases will introduce hashing of "
                        "artifacts which will improve this behavior."
                    )
            elif isinstance(value, LazyArtifactVersionResponse):
                model_artifacts_or_metadata[key] = ModelVersionDataLazyLoader(
                    model_name=value.lazy_load_model_name,
                    model_version=value.lazy_load_model_version,
                    artifact_name=value.lazy_load_name,
                    artifact_version=value.lazy_load_version,
                    metadata_name=None,
                )
            elif isinstance(value, ArtifactVersionResponse):
                external_artifacts[key] = value
            elif isinstance(value, LazyRunMetadataResponse):
                model_artifacts_or_metadata[key] = ModelVersionDataLazyLoader(
                    model_name=value.lazy_load_model_name,
                    model_version=value.lazy_load_model_version,
                    artifact_name=value.lazy_load_artifact_name,
                    artifact_version=value.lazy_load_artifact_version,
                    metadata_name=value.lazy_load_metadata_name,
                )
            elif isinstance(value, ClientLazyLoader):
                client_lazy_loaders[key] = value
            else:
                parameters[key] = value

        # Above we iterated over the provided arguments which should overwrite
        # any parameters previously defined on the step instance. Now we apply
        # the default values on the entrypoint function and add those as
        # parameters for any argument that has no value yet. If we were to do
        # that in the above loop, we would overwrite previously configured
        # parameters with the default values.
        bound_args.apply_defaults()
        for key, value in bound_args.arguments.items():
            self.entrypoint_definition.validate_input(key=key, value=value)
            if (
                key not in artifacts
                and key not in external_artifacts
                and key not in model_artifacts_or_metadata
                and key not in self.configuration.parameters
                and key not in client_lazy_loaders
            ):
                default_parameters[key] = value

        return (
            artifacts,
            external_artifacts,
            model_artifacts_or_metadata,
            client_lazy_loaders,
            parameters,
            default_parameters,
        )

    def _handle_call_during_compilation(
        self,
        compilation_context: "PipelineCompilationContext",
        args: Tuple[Any, ...],
        kwargs: Dict[str, Any],
        id: Optional[str] = None,
        after: Union[
            str,
            StepArtifact,
            Sequence[Union[str, StepArtifact]],
            None,
        ] = None,
    ) -> Union[StepArtifact, List[StepArtifact]]:
        """Handles a call during compilation of the step.

        Args:
            compilation_context: The compilation context.
            args: The arguments passed to the step.
            kwargs: The keyword arguments passed to the step.
            id: The ID of the step invocation.
            after: The upstream steps for the invocation

        Returns:
            The outputs of the step invocation.
        """
        (
            input_artifacts,
            external_artifacts,
            model_artifacts_or_metadata,
            client_lazy_loaders,
            parameters,
            default_parameters,
        ) = self._parse_call_args(*args, **kwargs)

        upstream_steps = {
            artifact.invocation_id
            for artifact_list in input_artifacts.values()
            for artifact in artifact_list
        }
        if isinstance(after, str):
            upstream_steps.add(after)
        elif isinstance(after, StepArtifact):
            upstream_steps.add(after.invocation_id)
        elif isinstance(after, Sequence):
            for item in after:
                if isinstance(item, str):
                    upstream_steps.add(item)
                elif isinstance(item, StepArtifact):
                    upstream_steps.add(item.invocation_id)

        invocation_id = compilation_context.pipeline.add_step_invocation(
            step=self,
            input_artifacts=input_artifacts,
            external_artifacts=external_artifacts,
            model_artifacts_or_metadata=model_artifacts_or_metadata,
            client_lazy_loaders=client_lazy_loaders,
            parameters=parameters,
            default_parameters=default_parameters,
            upstream_steps=upstream_steps,
            custom_id=id,
            allow_id_suffix=not id,
        )

        outputs = []
        for key, annotation in self.entrypoint_definition.outputs.items():
            output = StepArtifact(
                invocation_id=invocation_id,
                output_name=key,
                annotation=annotation,
                pipeline=compilation_context.pipeline,
            )
            outputs.append(output)
        return outputs[0] if len(outputs) == 1 else outputs

    def __call__(
        self,
        *args: Any,
        id: Optional[str] = None,
        after: Union[
            str,
            StepArtifact,
            "AnyStepFuture",
            Sequence[Union[str, StepArtifact, "AnyStepFuture"]],
            None,
        ] = None,
        **kwargs: Any,
    ) -> Any:
        """Handle a call of the step.

        This method does one of two things:
        * If there is an active pipeline context, it adds an invocation of the
          step instance to the pipeline.
        * If no pipeline is active, it calls the step entrypoint function.

        Args:
            *args: Entrypoint function arguments.
            id: Invocation ID to use.
            after: Upstream steps for the invocation.
            **kwargs: Entrypoint function keyword arguments.

        Returns:
            The outputs of the entrypoint function call.
        """
        from zenml import get_step_context
        from zenml.execution.pipeline.dynamic.run_context import (
            DynamicPipelineRunContext,
        )
        from zenml.execution.pipeline.utils import (
            should_prevent_pipeline_execution,
        )
        from zenml.pipelines.compilation_context import (
            PipelineCompilationContext,
        )

        compilation_context = PipelineCompilationContext.get()
        if compilation_context:
            # We're currently compiling a static pipeline, which we want to
            # allow even while inside a running step.
            after = cast(
                Union[
                    str,
                    "StepArtifact",
                    Sequence[Union[str, "StepArtifact"]],
                    None,
                ],
                after,
            )
            return self._handle_call_during_compilation(
                compilation_context=compilation_context,
                args=args,
                kwargs=kwargs,
                id=id,
                after=after,
            )

        try:
            step_context = get_step_context()
        except RuntimeError:
            step_context = None

        if step_context:
            # We're currently inside the execution of a different step
            # -> We don't want to launch another single step pipeline here,
            # but instead just call the step function
            return self.call_entrypoint(*args, **kwargs)

        if run_context := DynamicPipelineRunContext.get():
            after = cast(
                Union[
                    "AnyStepFuture",
                    Sequence["AnyStepFuture"],
                    None,
                ],
                after,
            )
            return run_context.runner.launch_step(
                step=self,
                id=id,
                args=args,
                kwargs=kwargs,
                after=after,
                concurrent=False,
            )

        # If the environment variable was set to explicitly not run on the
        # stack, we do that.
        run_without_stack = handle_bool_env_var(
            ENV_ZENML_RUN_SINGLE_STEPS_WITHOUT_STACK, default=False
        )
        if run_without_stack:
            return self.call_entrypoint(*args, **kwargs)

        if should_prevent_pipeline_execution():
            logger.info("Preventing execution of step '%s'.", self.name)
            return

        return run_as_single_step_pipeline(self, None, *args, **kwargs)

    def call_entrypoint(self, *args: Any, **kwargs: Any) -> Any:
        """Calls the entrypoint function of the step.

        Args:
            *args: Entrypoint function arguments.
            **kwargs: Entrypoint function keyword arguments.

        Returns:
            The return value of the entrypoint function.

        Raises:
            StepInterfaceError: If the arguments to the entrypoint function are
                invalid.
        """
        try:
            validated_args = pydantic_utils.validate_function_args(
                self.entrypoint,
                ConfigDict(arbitrary_types_allowed=True),
                *args,
                **kwargs,
            )
        except ValidationError as e:
            raise StepInterfaceError(
                "Invalid step function entrypoint arguments. Check out the "
                "pydantic error above for more details."
            ) from e

        return self.entrypoint(**validated_args)

    def submit(
        self,
        *args: Any,
        id: Optional[str] = None,
        after: Union["AnyStepFuture", Sequence["AnyStepFuture"], None] = None,
        **kwargs: Any,
    ) -> "StepFuture":
        """Submit the step to run concurrently in a separate thread.

        Args:
            *args: The arguments to pass to the step function.
            id: The invocation ID of the step.
            after: The step run output futures to wait for before executing the
                step.
            **kwargs: The keyword arguments to pass to the step function.

        Raises:
            RuntimeError: If this method is called outside of a dynamic
                pipeline.

        Returns:
            The step future.
        """
        from zenml.execution.pipeline.dynamic.run_context import (
            DynamicPipelineRunContext,
        )

        context = DynamicPipelineRunContext.get()
        if not context:
            raise RuntimeError(
                "Submitting a step is only possible within a dynamic pipeline."
            )

        return context.runner.launch_step(
            step=self,
            id=id,
            args=args,
            kwargs=kwargs,
            after=after,
            concurrent=True,
        )

    def map(
        self,
        *args: Any,
        after: Union["AnyStepFuture", Sequence["AnyStepFuture"], None] = None,
        **kwargs: Any,
    ) -> "MapResultsFuture":
        """Map over step inputs.

        This method will launch separate steps for each chunk of the input
        artifacts. This currently is only supported for inputs that are outputs
        from upstream steps in the same pipeline.

        Example:
        The following code will launch 2 `do_something` steps, which will
        receive the inputs `(1, "a")` and `(2, "b")`.

        ```python
        from zenml import pipeline, step

        @step
        def create_int_list() -> list[int]:
            return [1, 2]

        @step
        def create_str_list() -> list[str]:
            return ["a", "b"]

        @step
        def do_something(a: int, b: str) -> None:
            ...

        @pipeline
        def map_pipeline():
            int_list = create_int_list()
            str_list = create_str_list()
            do_something.map(a=int_list, b=str_list)
        ```

        Args:
            *args: The arguments to pass to the step function.
            after: The step run output futures to wait for before executing the
                steps.
            **kwargs: The keyword arguments to pass to the step function.

        Raises:
            RuntimeError: If this method is called outside of a dynamic
                pipeline.

        Returns:
            A future that represents the map results.
        """
        from zenml.execution.pipeline.dynamic.run_context import (
            DynamicPipelineRunContext,
        )

        context = DynamicPipelineRunContext.get()
        if not context:
            raise RuntimeError(
                "Mapping over step inputs is only possible within a dynamic "
                "pipeline."
            )

        return context.runner.map(
            step=self,
            args=args,
            kwargs=kwargs,
            after=after,
            product=False,
        )

    def product(
        self,
        *args: Any,
        after: Union["AnyStepFuture", Sequence["AnyStepFuture"], None] = None,
        **kwargs: Any,
    ) -> "MapResultsFuture":
        """Map over step inputs using a cartesian product of the mapped inputs.

        This method will launch separate steps for each combination of the
        iteams of the input artifacts. This currently is only supported for
        inputs that are outputs from upstream steps in the same pipeline.

        Example:
        The following code will launch 4 `do_something` steps, which will
        receive the inputs `(1, "a")`, `(1, "b")`, `(2, "a")`, `(2, "b")`.

        ```python
        from zenml import pipeline, step

        @step
        def create_int_list() -> list[int]:
            return [1, 2]

        @step
        def create_str_list() -> list[str]:
            return ["a", "b"]

        @step
        def do_something(a: int, b: str) -> None:
            ...

        @pipeline
        def map_pipeline():
            int_list = create_int_list()
            str_list = create_str_list()
            do_something.product(a=int_list, b=str_list)
        ```

        Args:
            *args: The arguments to pass to the step function.
            after: The step run output futures to wait for before executing the
                steps.
            **kwargs: The keyword arguments to pass to the step function.

        Raises:
            RuntimeError: If this method is called outside of a dynamic
                pipeline.

        Returns:
            A future that represents the map results.
        """
        from zenml.execution.pipeline.dynamic.run_context import (
            DynamicPipelineRunContext,
        )

        context = DynamicPipelineRunContext.get()
        if not context:
            raise RuntimeError(
                "Mapping over step inputs is only possible within a dynamic "
                "pipeline."
            )

        return context.runner.map(
            step=self,
            args=args,
            kwargs=kwargs,
            after=after,
            product=True,
        )

    @property
    def name(self) -> str:
        """The name of the step.

        Returns:
            The name of the step.
        """
        return self.configuration.name

    @property
    def enable_cache(self) -> Optional[bool]:
        """If caching is enabled for the step.

        Returns:
            If caching is enabled for the step.
        """
        return self.configuration.enable_cache

    @property
    def configuration(self) -> "PartialStepConfiguration":
        """The configuration of the step.

        Returns:
            The configuration of the step.
        """
        return self._configuration

    def configure(
        self: T,
        enable_cache: Optional[bool] = None,
        enable_artifact_metadata: Optional[bool] = None,
        enable_artifact_visualization: Optional[bool] = None,
        enable_step_logs: Optional[bool] = None,
        experiment_tracker: Optional[Union[bool, str]] = None,
        step_operator: Optional[Union[bool, str]] = None,
        parameters: Optional[Dict[str, Any]] = None,
        output_materializers: Optional[
            "OutputMaterializersSpecification"
        ] = None,
        environment: Optional[Dict[str, Any]] = None,
        secrets: Optional[Sequence[Union[str, UUID]]] = None,
        settings: Optional[Mapping[str, "SettingsOrDict"]] = None,
        extra: Optional[Dict[str, Any]] = None,
        on_failure: Optional["HookSpecification"] = None,
        on_success: Optional["HookSpecification"] = None,
        model: Optional["Model"] = None,
        retry: Optional[StepRetryConfig] = None,
        substitutions: Optional[Dict[str, str]] = None,
        cache_policy: Optional[CachePolicyOrString] = None,
        runtime: Optional[StepRuntime] = None,
        group: Optional[Union["GroupInfo", str]] = None,
        heartbeat_healthy_threshold: Optional[int] = None,
        merge: bool = True,
    ) -> T:
        """Configures the step.

        Configuration merging example:
        * `merge==True`:
            step.configure(extra={"key1": 1})
            step.configure(extra={"key2": 2}, merge=True)
            step.configuration.extra # {"key1": 1, "key2": 2}
        * `merge==False`:
            step.configure(extra={"key1": 1})
            step.configure(extra={"key2": 2}, merge=False)
            step.configuration.extra # {"key2": 2}

        Args:
            enable_cache: If caching should be enabled for this step.
            enable_artifact_metadata: If artifact metadata should be enabled
                for this step.
            enable_artifact_visualization: If artifact visualization should be
                enabled for this step.
            enable_step_logs: If step logs should be enabled for this step.
            experiment_tracker: The experiment tracker to use for this step.
            step_operator: The step operator to use for this step.
            parameters: Function parameters for this step
            output_materializers: Output materializers for this step. If
                given as a dict, the keys must be a subset of the output names
                of this step. If a single value (type or string) is given, the
                materializer will be used for all outputs.
            environment: Environment variables to set when running this step.
            secrets: Secrets to set as environment variables when running this
                step.
            settings: Settings for this step.
            extra: Extra configurations for this step.
            on_failure: Callback function in event of failure of the step. Can
                be a function with a single argument of type `BaseException`, or
                a source path to such a function (e.g. `module.my_function`).
            on_success: Callback function in event of success of the step. Can
                be a function with no arguments, or a source path to such a
                function (e.g. `module.my_function`).
            model: Model to use for this step.
            retry: Configuration for retrying the step in case of failure.
            substitutions: Extra substitutions for model and artifact name
                placeholders.
            cache_policy: Cache policy for this step.
            runtime: The step runtime. This is only applicable for dynamic
                pipelines.
            heartbeat_healthy_threshold: The amount of time (in minutes) that a
                running step has not received heartbeat and is considered healthy.
                By default, set to 30 minutes.
            group: The group information for this step.
            merge: If `True`, will merge the given dictionary configurations
                like `parameters` and `settings` with existing
                configurations. If `False` the given configurations will
                overwrite all existing ones. See the general description of this
                method for an example.

        Returns:
            The step instance that this method was called on.
        """
        from zenml.config.step_configurations import StepConfigurationUpdate
        from zenml.hooks.hook_validators import resolve_and_validate_hook

        def _resolve_if_necessary(
            value: Union[str, Source, Type[Any]],
        ) -> Source:
            if isinstance(value, str):
                return Source.from_import_path(value)
            elif isinstance(value, Source):
                return value
            else:
                return source_utils.resolve(value)

        def _convert_to_tuple(value: Any) -> Tuple[Source, ...]:
            if isinstance(value, str) or not isinstance(value, Sequence):
                return (_resolve_if_necessary(value),)
            else:
                return tuple(_resolve_if_necessary(v) for v in value)

        outputs: Dict[str, Dict[str, Tuple[Source, ...]]] = defaultdict(dict)
        allowed_output_names = set(self.entrypoint_definition.outputs)

        if output_materializers:
            if not isinstance(output_materializers, Mapping):
                sources = _convert_to_tuple(output_materializers)
                output_materializers = {
                    output_name: sources
                    for output_name in allowed_output_names
                }

            for output_name, materializer in output_materializers.items():
                sources = _convert_to_tuple(materializer)
                outputs[output_name]["materializer_source"] = sources

        failure_hook_source = None
        if on_failure:
            # string of on_failure hook function to be used for this step
            failure_hook_source, _ = resolve_and_validate_hook(
                on_failure, allow_exception_arg=True
            )

        success_hook_source = None
        if on_success:
            # string of on_success hook function to be used for this step
            success_hook_source, _ = resolve_and_validate_hook(on_success)

        if merge and secrets and self._configuration.secrets:
            secrets = self._configuration.secrets + list(secrets)

        if isinstance(group, str):
            group = GroupInfo(id=group, name=group, type=GroupType.MANUAL)

        values = dict_utils.remove_none_values(
            {
                "enable_cache": enable_cache,
                "enable_artifact_metadata": enable_artifact_metadata,
                "enable_artifact_visualization": enable_artifact_visualization,
                "enable_step_logs": enable_step_logs,
                "experiment_tracker": experiment_tracker,
                "step_operator": step_operator,
                "parameters": parameters,
                "environment": environment,
                "secrets": secrets,
                "settings": settings,
                "outputs": outputs or None,
                "extra": extra,
                "failure_hook_source": failure_hook_source,
                "success_hook_source": success_hook_source,
                "model": model,
                "retry": retry,
                "substitutions": substitutions,
                "cache_policy": cache_policy,
                "runtime": runtime,
                "heartbeat_healthy_threshold": heartbeat_healthy_threshold,
                "group": group,
            }
        )
        config = StepConfigurationUpdate(**values)
        self._apply_configuration(config, merge=merge)
        return self

    def with_options(
        self,
        enable_cache: Optional[bool] = None,
        enable_artifact_metadata: Optional[bool] = None,
        enable_artifact_visualization: Optional[bool] = None,
        enable_step_logs: Optional[bool] = None,
        experiment_tracker: Optional[Union[bool, str]] = None,
        step_operator: Optional[Union[bool, str]] = None,
        parameters: Optional[Dict[str, Any]] = None,
        output_materializers: Optional[
            "OutputMaterializersSpecification"
        ] = None,
        environment: Optional[Dict[str, Any]] = None,
        secrets: Optional[List[Union[str, UUID]]] = None,
        settings: Optional[Mapping[str, "SettingsOrDict"]] = None,
        extra: Optional[Dict[str, Any]] = None,
        on_failure: Optional["HookSpecification"] = None,
        on_success: Optional["HookSpecification"] = None,
        model: Optional["Model"] = None,
        retry: Optional[StepRetryConfig] = None,
        substitutions: Optional[Dict[str, str]] = None,
        cache_policy: Optional[CachePolicyOrString] = None,
        runtime: Optional[StepRuntime] = None,
        heartbeat_healthy_threshold: Optional[int] = None,
        group: Optional[Union["GroupInfo", str]] = None,
        merge: bool = True,
    ) -> "BaseStep":
        """Copies the step and applies the given configurations.

        Args:
            enable_cache: If caching should be enabled for this step.
            enable_artifact_metadata: If artifact metadata should be enabled
                for this step.
            enable_artifact_visualization: If artifact visualization should be
                enabled for this step.
            enable_step_logs: If step logs should be enabled for this step.
            experiment_tracker: The experiment tracker to use for this step.
            step_operator: The step operator to use for this step.
            parameters: Function parameters for this step
            output_materializers: Output materializers for this step. If
                given as a dict, the keys must be a subset of the output names
                of this step. If a single value (type or string) is given, the
                materializer will be used for all outputs.
            environment: Environment variables to set when running this step.
            secrets: Secrets to set as environment variables when running this
                step.
            settings: Settings for this step.
            extra: Extra configurations for this step.
            on_failure: Callback function in event of failure of the step. Can
                be a function with a single argument of type `BaseException`, or
                a source path to such a function (e.g. `module.my_function`).
            on_success: Callback function in event of success of the step. Can
                be a function with no arguments, or a source path to such a
                function (e.g. `module.my_function`).
            model: Model to use for this step.
            retry: Configuration for retrying the step in case of failure.
            substitutions: Extra substitutions for model and artifact name
                placeholders.
            cache_policy: Cache policy for this step.
            runtime: The step runtime. This is only applicable for dynamic
                pipelines.
            heartbeat_healthy_threshold: The amount of time (in minutes) that a
                running step has not received heartbeat and is considered healthy.
                By default, set to 30 minutes.
            group: The group information for this step.
            merge: If `True`, will merge the given dictionary configurations
                like `parameters` and `settings` with existing
                configurations. If `False` the given configurations will
                overwrite all existing ones. See the general description of this
                method for an example.

        Returns:
            The copied step instance.
        """
        step_copy = self.copy()
        step_copy.configure(
            enable_cache=enable_cache,
            enable_artifact_metadata=enable_artifact_metadata,
            enable_artifact_visualization=enable_artifact_visualization,
            enable_step_logs=enable_step_logs,
            experiment_tracker=experiment_tracker,
            step_operator=step_operator,
            parameters=parameters,
            output_materializers=output_materializers,
            environment=environment,
            secrets=secrets,
            settings=settings,
            extra=extra,
            on_failure=on_failure,
            on_success=on_success,
            model=model,
            retry=retry,
            substitutions=substitutions,
            cache_policy=cache_policy,
            runtime=runtime,
            heartbeat_healthy_threshold=heartbeat_healthy_threshold,
            group=group,
            merge=merge,
        )
        return step_copy

    def copy(self) -> "BaseStep":
        """Copies the step.

        Returns:
            The step copy.
        """
        from zenml.execution.pipeline.dynamic.run_context import (
            DynamicPipelineRunContext,
        )

        step_copy = copy.deepcopy(self)

        if not DynamicPipelineRunContext.is_active():
            # If we're not in a dynamic pipeline, we generate a new static ID
            # for the step copy
            step_copy._static_id = id(step_copy)

        return step_copy

    @contextmanager
    def _suspend_dynamic_configuration(self) -> Generator[None, None, None]:
        """Context manager to suspend applying to the dynamic configuration.

        Yields:
            None.
        """
        previous_value = self._capture_dynamic_configuration
        self._capture_dynamic_configuration = False
        try:
            yield
        finally:
            self._capture_dynamic_configuration = previous_value

    def _apply_configuration(
        self,
        config: "StepConfigurationUpdate",
        merge: bool = True,
        runtime_parameters: Dict[str, Any] = {},
    ) -> None:
        """Applies an update to the step configuration.

        Args:
            config: The configuration update.
            runtime_parameters: Dictionary of parameters passed to a step from runtime
            merge: Whether to merge the updates with the existing configuration
                or not. See the `BaseStep.configure(...)` method for a detailed
                explanation.
        """
        from zenml.execution.pipeline.dynamic.run_context import (
            DynamicPipelineRunContext,
        )

        self._validate_configuration(config, runtime_parameters)

        if (
            self._capture_dynamic_configuration
            and DynamicPipelineRunContext.is_active()
        ):
            if self._dynamic_configuration is None:
                self._dynamic_configuration = config
            else:
                self._dynamic_configuration = pydantic_utils.update_model(
                    self._dynamic_configuration, update=config, recursive=merge
                )
            return

        self._configuration = pydantic_utils.update_model(
            self._configuration, update=config, recursive=merge
        )

        logger.debug("Updated step configuration:")
        logger.debug(self._configuration)

    def _merge_dynamic_configuration(self) -> None:
        """Merges the dynamic configuration into the static configuration."""
        if self._dynamic_configuration:
            with self._suspend_dynamic_configuration():
                self._apply_configuration(
                    config=self._dynamic_configuration, merge=True
                )
            logger.debug("Merged dynamic configuration.")

    def _validate_configuration(
        self,
        config: "StepConfigurationUpdate",
        runtime_parameters: Dict[str, Any],
    ) -> None:
        """Validates a configuration update.

        Args:
            config: The configuration update to validate.
            runtime_parameters: Dictionary of parameters passed to a step from runtime
        """
        if config.settings:
            settings_utils.validate_setting_keys(list(config.settings))
        self._validate_function_parameters(
            parameters=config.parameters, runtime_parameters=runtime_parameters
        )
        self._validate_outputs(outputs=config.outputs)

    def _validate_function_parameters(
        self,
        parameters: Optional[Dict[str, Any]],
        runtime_parameters: Dict[str, Any],
    ) -> None:
        """Validates step function parameters.

        Args:
            parameters: The parameters to validate.
            runtime_parameters: Dictionary of parameters passed to a step from runtime

        Raises:
            StepInterfaceError: If the step requires no function parameters but
                parameters were configured.
            RuntimeError: If the step has parameters configured differently in
                configuration file and code.
        """
        if not parameters:
            return

        conflicting_parameters = {}
        for key, value in parameters.items():
            if key in runtime_parameters:
                runtime_value = runtime_parameters[key]
                if runtime_value != value:
                    conflicting_parameters[key] = (value, runtime_value)
            if key in self.entrypoint_definition.inputs:
                self.entrypoint_definition.validate_input(key=key, value=value)
            else:
                raise StepInterfaceError(
                    f"Unable to find parameter '{key}' in step function "
                    "signature."
                )
        if conflicting_parameters:
            is_plural = "s" if len(conflicting_parameters) > 1 else ""
            msg = f"Configured parameter{is_plural} for the step '{self.name}' conflict{'' if not is_plural else 's'} with parameter{is_plural} passed in runtime:\n"
            for key, values in conflicting_parameters.items():
                msg += (
                    f"`{key}`: config=`{values[0]}` | runtime=`{values[1]}`\n"
                )
            msg += """This happens, if you define values for step parameters in configuration file and pass same parameters from the code. Example:
```
# config.yaml

steps:
    step_name:
        parameters:
            param_name: value1
            
            
# pipeline.py

@pipeline
def pipeline_():
    step_name(param_name="other_value")
```
To avoid this consider setting step parameters only in one place (config or code).
"""
            raise RuntimeError(msg)

    def _validate_outputs(
        self, outputs: Mapping[str, "PartialArtifactConfiguration"]
    ) -> None:
        """Validates the step output configuration.

        Args:
            outputs: The configured step outputs.

        Raises:
            StepInterfaceError: If an output for a non-existent name is
                configured of an output artifact/materializer source does not
                resolve to the correct class.
        """
        allowed_output_names = set(self.entrypoint_definition.outputs)
        for output_name, output in outputs.items():
            if output_name not in allowed_output_names:
                raise StepInterfaceError(
                    f"Got unexpected materializers for non-existent "
                    f"output '{output_name}' in step '{self.name}'. "
                    f"Only materializers for the outputs "
                    f"{allowed_output_names} of this step can"
                    f" be registered."
                )

            if output.materializer_source:
                for source in output.materializer_source:
                    try:
                        source_utils.validate_source_class(
                            source, expected_class=BaseMaterializer
                        )
                    except SourceValidationException as e:
                        raise StepInterfaceError(
                            f"Materializer source `{source}` "
                            f"for output '{output_name}' of step '{self.name}' "
                            "does not resolve to a `BaseMaterializer` subclass."
                        ) from e

    def _validate_inputs(
        self,
        input_artifacts: Dict[str, List["StepArtifact"]],
        external_artifacts: Dict[str, "ExternalArtifactConfiguration"],
        model_artifacts_or_metadata: Dict[str, "ModelVersionDataLazyLoader"],
        client_lazy_loaders: Dict[str, "ClientLazyLoader"],
    ) -> None:
        """Validates the step inputs.

        This method makes sure that all inputs are provided either as an
        artifact or parameter.

        Args:
            input_artifacts: The input artifacts.
            external_artifacts: The external input artifacts.
            model_artifacts_or_metadata: The model artifacts or metadata.
            client_lazy_loaders: The client lazy loaders.

        Raises:
            StepInterfaceError: If an entrypoint input is missing.
        """
        for key in self.entrypoint_definition.inputs.keys():
            if (
                key in input_artifacts
                or key in self.configuration.parameters
                or key in external_artifacts
                or key in model_artifacts_or_metadata
                or key in client_lazy_loaders
            ):
                continue
            raise StepInterfaceError(
                f"Missing entrypoint input '{key}' in step '{self.name}'."
            )

    def _finalize_configuration(
        self,
        input_artifacts: Dict[str, List["StepArtifact"]],
        external_artifacts: Dict[str, "ExternalArtifactConfiguration"],
        model_artifacts_or_metadata: Dict[str, "ModelVersionDataLazyLoader"],
        client_lazy_loaders: Dict[str, "ClientLazyLoader"],
        skip_input_validation: bool = False,
    ) -> "StepConfiguration":
        """Finalizes the configuration after the step was called.

        Once the step was called, we know the outputs of previous steps
        and that no additional user configurations will be made. That means
        we can now collect the remaining artifact and materializer types
        as well as check for the completeness of the step function parameters.

        Args:
            input_artifacts: The input artifacts of this step.
            external_artifacts: The external artifacts of this step.
            model_artifacts_or_metadata: The model artifacts or metadata of
                this step.
            client_lazy_loaders: The client lazy loaders of this step.
            skip_input_validation: If True, will skip the input validation.

        Raises:
            StepInterfaceError: If explicit materializers were specified for an
                output but they do not work for the data type(s) defined by
                the type annotation.

        Returns:
            The finalized step configuration.
        """
        from zenml.config.step_configurations import (
            PartialArtifactConfiguration,
            StepConfiguration,
            StepConfigurationUpdate,
        )

        outputs: Dict[str, Dict[str, Any]] = defaultdict(dict)

        for (
            output_name,
            output_annotation,
        ) in self.entrypoint_definition.outputs.items():
            output = self._configuration.outputs.get(
                output_name, PartialArtifactConfiguration()
            )
            if artifact_config := output_annotation.artifact_config:
                outputs[output_name]["artifact_config"] = artifact_config

            if output.materializer_source:
                # The materializer source was configured by the user. We
                # validate that their configured materializer supports the
                # output type. If the output annotation is a Union, we check
                # that at least one of the specified materializers works with at
                # least one of the types in the Union. If that's not the case,
                # it would be a guaranteed failure at runtime and we fail early
                # here.
                if output_annotation.resolved_annotation is Any:
                    continue

                materializer_classes: List[Type["BaseMaterializer"]] = [
                    source_utils.load(materializer_source)
                    for materializer_source in output.materializer_source
                ]

                for data_type in output_annotation.get_output_types():
                    try:
                        materializer_utils.select_materializer(
                            data_type=data_type,
                            materializer_classes=materializer_classes,
                        )
                        break
                    except RuntimeError:
                        pass
                else:
                    materializer_strings = [
                        materializer_source.import_path
                        for materializer_source in output.materializer_source
                    ]
                    raise StepInterfaceError(
                        "Invalid materializers specified for output "
                        f"{output_name} of step {self.name}. None of the "
                        f"materializers ({materializer_strings}) are "
                        "able to save or load data of the type that is defined "
                        "for the output "
                        f"({output_annotation.resolved_annotation})."
                    )
            else:
                if output_annotation.resolved_annotation is Any:
                    outputs[output_name]["materializer_source"] = ()
                    outputs[output_name]["default_materializer_source"] = (
                        source_utils.resolve(
                            materializer_registry.get_default_materializer()
                        )
                    )
                    continue

                materializer_sources = []

                for output_type in output_annotation.get_output_types():
                    materializer_class = materializer_registry[output_type]
                    materializer_sources.append(
                        source_utils.resolve(materializer_class)
                    )

                outputs[output_name]["materializer_source"] = tuple(
                    materializer_sources
                )

        parameters = self._finalize_parameters()
        self.configure(parameters=parameters, merge=False)
        if not skip_input_validation:
            self._validate_inputs(
                input_artifacts=input_artifacts,
                external_artifacts=external_artifacts,
                model_artifacts_or_metadata=model_artifacts_or_metadata,
                client_lazy_loaders=client_lazy_loaders,
            )

        values = dict_utils.remove_none_values({"outputs": outputs or None})
        config = StepConfigurationUpdate(**values)
        self._apply_configuration(config)

        self._configuration = self._configuration.model_copy(
            update={
                "caching_parameters": self.caching_parameters,
                "external_input_artifacts": external_artifacts,
                "model_artifacts_or_metadata": model_artifacts_or_metadata,
                "client_lazy_loaders": client_lazy_loaders,
            }
        )

        return StepConfiguration.model_validate(
            self._configuration.model_dump()
        )

    def _finalize_parameters(self) -> Dict[str, Any]:
        """Finalizes the config parameters for running this step.

        Returns:
            All parameter values for running this step.
        """
        params = {}
        for key, value in self.configuration.parameters.items():
            if key not in self.entrypoint_definition.inputs:
                continue

            annotation = self.entrypoint_definition.inputs[key].annotation
            annotation = resolve_type_annotation(annotation)
            if inspect.isclass(annotation) and issubclass(
                annotation, BaseModel
            ):
                # Make sure we have all necessary values to instantiate the
                # pydantic model later
                model = annotation(**value)
                params[key] = model.model_dump()
            else:
                params[key] = value

        return params

    def _compute_parameter_schema(self) -> Optional[Dict[str, Any]]:
        """Computes a JSON schema for the configured step parameters.

        Returns:
            The JSON schema for the configured step parameters.
        """
        parameter_names = set(self.configuration.parameters)
        if not parameter_names:
            return {}

        parameter_inputs = {
            name: parameter
            for name, parameter in self.entrypoint_definition.inputs.items()
            if name in parameter_names
        }

        try:
            parameter_model = pydantic_utils.create_parameter_model(
                model_name=f"{self.name.title().replace('_', '')}Parameters",
                parameters=parameter_inputs,
                default_values=self.configuration.parameters,
            )
            return parameter_model.model_json_schema()
        except Exception as e:
            logger.debug(
                "Failed to generate the parameter schema for step `%s`: %s.",
                self.name,
                e,
            )
            return None

    def replay(
        self,
        pipeline: Union[UUID, str, None] = None,
        pipeline_run: Union[UUID, str, None] = None,
        step_run_id: Optional[UUID] = None,
        invocation_id: Optional[str] = None,
        input_overrides: Optional[Mapping[str, Any]] = None,
        debug: bool = False,
    ) -> Any:
        """Replay the step.

        The step to replay gets determined as follows:
        - If you specify a step run ID, that specific step run is replayed.
        - If you specify a pipeline run, the step run with the given invocation
          ID (or the step name if no invocation ID is specified) is replayed.
        - If you specify a pipeline, the last run of that pipeline is used. The
          step run with the given invocation ID (or the step name if no
          invocation ID is specified) is replayed.

        Args:
            pipeline: The pipeline to replay.
            pipeline_run: The pipeline run to replay.
            step_run_id: The step run ID to replay.
            invocation_id: The invocation ID of the step to replay.
            input_overrides: Input overrides for the step.
            debug: Whether to run the step in debug mode. In debug mode, the
                step is executed using a local orchestrator, while keeping the
                remaining components of your active stack.

        Raises:
            ValueError: If no step run can be found.

        Returns:
            The outputs of the replayed step run.
        """
        from zenml.client import Client
        from zenml.execution.utils import DebugModeContext

        invocation_id = invocation_id or self.name

        if step_run_id:
            step_run = Client().get_run_step(step_run_id)
        elif pipeline_run:
            pipeline_run_model = Client().get_pipeline_run(pipeline_run)
            step_runs = Client().list_run_steps(
                pipeline_run_id=pipeline_run_model.id, name=invocation_id
            )
            if len(step_runs) == 0:
                raise ValueError(
                    f"No existing step run found for step `{invocation_id}` in "
                    f"pipeline run `{pipeline_run}`."
                )
            step_run = step_runs[0]
        elif pipeline:
            pipeline_model = Client().get_pipeline(pipeline)
            step_runs = Client().list_run_steps(
                pipeline_run_id=pipeline_model.last_run.id, name=invocation_id
            )
            if len(step_runs) == 0:
                raise ValueError(
                    f"No existing step run found for step `{invocation_id}` "
                    f"in latest run of pipeline `{pipeline_model.name}`."
                )
            step_run = step_runs[0]
        else:
            raise ValueError(
                "No pipeline, pipeline run, or step run provided."
            )

        if step_run.code_hash != self.source_code_cache_value:
            logger.warning(
                "The code of the step run that you're trying to replay does "
                "not match your local step code."
            )

        inputs = {}
        for input_name, input_artifacts in step_run.regular_inputs.items():
            if len(input_artifacts) > 1:
                inputs[input_name] = [
                    artifact.load() for artifact in input_artifacts
                ]
            else:
                inputs[input_name] = input_artifacts[0].load()

        if input_overrides:
            inputs.update(input_overrides)

        step_instance = self.copy()

        if debug:
            with DebugModeContext():
                return run_as_single_step_pipeline(
                    step_instance, f"{self.name}_replay", **inputs
                )
        else:
            return run_as_single_step_pipeline(
                step_instance, f"{self.name}_replay", **inputs
            )



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Implementation of the ZenML local Docker orchestrator."""

import copy
import os
import sys
import time
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Type, cast
from uuid import uuid4

from docker.errors import ContainerError
from pydantic import Field

from zenml.config.base_settings import BaseSettings
from zenml.config.global_config import GlobalConfiguration
from zenml.constants import (
    ENV_ZENML_LOCAL_STORES_PATH,
)
from zenml.entrypoints import StepEntrypointConfiguration
from zenml.enums import ExecutionMode, StackComponentType
from zenml.logger import get_logger
from zenml.orchestrators import (
    BaseOrchestratorConfig,
    BaseOrchestratorFlavor,
    ContainerizedOrchestrator,
    SubmissionResult,
)
from zenml.stack import Stack, StackValidator
from zenml.utils import docker_utils, string_utils

if TYPE_CHECKING:
    from zenml.models import PipelineRunResponse, PipelineSnapshotResponse

logger = get_logger(__name__)

ENV_ZENML_DOCKER_ORCHESTRATOR_RUN_ID = "ZENML_DOCKER_ORCHESTRATOR_RUN_ID"


class LocalDockerOrchestrator(ContainerizedOrchestrator):
    """Orchestrator responsible for running pipelines locally using Docker.

    This orchestrator does not allow for concurrent execution of steps and also
    does not support running on a schedule.
    """

    @property
    def settings_class(self) -> Optional[Type["BaseSettings"]]:
        """Settings class for the Local Docker orchestrator.

        Returns:
            The settings class.
        """
        return LocalDockerOrchestratorSettings

    @property
    def config(self) -> "LocalDockerOrchestratorConfig":
        """Returns the `LocalDockerOrchestratorConfig` config.

        Returns:
            The configuration.
        """
        return cast(LocalDockerOrchestratorConfig, self._config)

    @property
    def validator(self) -> Optional[StackValidator]:
        """Ensures there is an image builder in the stack.

        Returns:
            A `StackValidator` instance.
        """
        return StackValidator(
            required_components={StackComponentType.IMAGE_BUILDER}
        )

    def get_orchestrator_run_id(self) -> str:
        """Returns the active orchestrator run id.

        Raises:
            RuntimeError: If the environment variable specifying the run id
                is not set.

        Returns:
            The orchestrator run id.
        """
        try:
            return os.environ[ENV_ZENML_DOCKER_ORCHESTRATOR_RUN_ID]
        except KeyError:
            raise RuntimeError(
                "Unable to read run id from environment variable "
                f"{ENV_ZENML_DOCKER_ORCHESTRATOR_RUN_ID}."
            )

    def submit_pipeline(
        self,
        snapshot: "PipelineSnapshotResponse",
        stack: "Stack",
        base_environment: Dict[str, str],
        step_environments: Dict[str, Dict[str, str]],
        placeholder_run: Optional["PipelineRunResponse"] = None,
    ) -> Optional[SubmissionResult]:
        """Submits a pipeline to the orchestrator.

        This method should only submit the pipeline and not wait for it to
        complete. If the orchestrator is configured to wait for the pipeline run
        to complete, a function that waits for the pipeline run to complete can
        be passed as part of the submission result.

        Args:
            snapshot: The pipeline snapshot to submit.
            stack: The stack the pipeline will run on.
            base_environment: Base environment shared by all steps. This should
                be set if your orchestrator for example runs one container that
                is responsible for starting all the steps.
            step_environments: Environment variables to set when executing
                specific steps.
            placeholder_run: An optional placeholder run for the snapshot.

        Raises:
            ContainerError: If the pipeline run fails.
            RuntimeError: If the pipeline run fails.

        Returns:
            Optional submission result.
        """
        if snapshot.schedule:
            logger.warning(
                "Local Docker Orchestrator currently does not support the "
                "use of schedules. The `schedule` will be ignored "
                "and the pipeline will be run immediately."
            )

        docker_client = docker_utils._try_get_docker_client_from_env()

        entrypoint = StepEntrypointConfiguration.get_entrypoint_command()

        # Add the local stores path as a volume mount
        stack.check_local_paths()
        local_stores_path = GlobalConfiguration().local_stores_path
        volumes = {
            local_stores_path: {
                "bind": local_stores_path,
                "mode": "rw",
            }
        }
        orchestrator_run_id = str(uuid4())
        start_time = time.time()

        execution_mode = snapshot.pipeline_configuration.execution_mode

        failed_steps: List[str] = []
        skipped_steps: List[str] = []

        # Run each step
        for step_name, step in snapshot.step_configurations.items():
            if (
                execution_mode == ExecutionMode.STOP_ON_FAILURE
                and failed_steps
            ):
                logger.warning(
                    "Skipping step %s due to the failed step(s): %s (Execution mode %s)",
                    step_name,
                    ", ".join(failed_steps),
                    execution_mode,
                )
                skipped_steps.append(step_name)
                continue

            if failed_upstream_steps := [
                fs for fs in failed_steps if fs in step.spec.upstream_steps
            ]:
                logger.warning(
                    "Skipping step %s due to failure in upstream step(s): %s (Execution mode %s)",
                    step_name,
                    ", ".join(failed_upstream_steps),
                    execution_mode,
                )
                skipped_steps.append(step_name)
                continue

            if skipped_upstream_steps := [
                fs for fs in skipped_steps if fs in step.spec.upstream_steps
            ]:
                logger.warning(
                    "Skipping step %s due to the skipped upstream step(s) %s (Execution mode %s)",
                    step_name,
                    ", ".join(skipped_upstream_steps),
                    execution_mode,
                )
                skipped_steps.append(step_name)
                continue

            if self.requires_resources_in_orchestration_environment(step):
                logger.warning(
                    "Specifying step resources is not supported for the local "
                    "Docker orchestrator, ignoring resource configuration for "
                    "step %s.",
                    step_name,
                )

            step_environment = step_environments[step_name]
            step_environment[ENV_ZENML_DOCKER_ORCHESTRATOR_RUN_ID] = (
                orchestrator_run_id
            )
            step_environment[ENV_ZENML_LOCAL_STORES_PATH] = local_stores_path

            arguments = StepEntrypointConfiguration.get_entrypoint_arguments(
                step_name=step_name, snapshot_id=snapshot.id
            )

            settings = cast(
                LocalDockerOrchestratorSettings,
                self.get_settings(step),
            )
            image = self.get_image(snapshot=snapshot, step_name=step_name)
            image = settings.run_args.pop("image", image)

            user = None
            if sys.platform != "win32":
                user = os.getuid()
            user = settings.run_args.pop("user", user)
            logger.info("Running step `%s` in Docker:", step_name)

            run_args = copy.deepcopy(settings.run_args)
            docker_environment = run_args.pop("environment", {})
            docker_environment.update(step_environment)

            docker_volumes = run_args.pop("volumes", {})
            docker_volumes.update(volumes)

            extra_hosts = run_args.pop("extra_hosts", {})
            extra_hosts["host.docker.internal"] = "host-gateway"

            try:
                logs = docker_client.containers.run(
                    image=image,
                    entrypoint=entrypoint,
                    command=arguments,
                    user=user,
                    volumes=docker_volumes,
                    environment=docker_environment,
                    stream=True,
                    extra_hosts=extra_hosts,
                    **run_args,
                )

                for line in logs:
                    logger.info(line.strip().decode())
            except ContainerError as e:
                error_message = e.stderr.decode()
                failed_steps.append(step_name)
                if execution_mode == ExecutionMode.FAIL_FAST:
                    raise
                else:
                    logger.error(error_message)

        if failed_steps:
            raise RuntimeError(
                "Pipeline run has failed due to failure in step(s): "
                f"{', '.join(failed_steps)}"
            )

        run_duration = time.time() - start_time
        logger.info(
            "Pipeline run has finished in `%s`.",
            string_utils.get_human_readable_time(run_duration),
        )
        return None

    def submit_dynamic_pipeline(
        self,
        snapshot: "PipelineSnapshotResponse",
        stack: "Stack",
        environment: Dict[str, str],
        placeholder_run: Optional["PipelineRunResponse"] = None,
    ) -> Optional[SubmissionResult]:
        """Submits a dynamic pipeline to the orchestrator.

        Args:
            snapshot: The pipeline snapshot to submit.
            stack: The stack the pipeline will run on.
            environment: Environment variables to set in the orchestration
                environment.
            placeholder_run: An optional placeholder run.

        Raises:
            RuntimeError: If the pipeline run fails.

        Returns:
            Optional submission result.
        """
        from zenml.pipelines.dynamic.entrypoint_configuration import (
            DynamicPipelineEntrypointConfiguration,
        )

        docker_client = docker_utils._try_get_docker_client_from_env()

        settings = cast(
            LocalDockerOrchestratorSettings,
            self.get_settings(snapshot),
        )

        command = (
            DynamicPipelineEntrypointConfiguration.get_entrypoint_command()
        )
        arguments = (
            DynamicPipelineEntrypointConfiguration.get_entrypoint_arguments(
                snapshot_id=snapshot.id,
                run_id=placeholder_run.id if placeholder_run else None,
            )
        )
        stack.check_local_paths()
        local_stores_path = GlobalConfiguration().local_stores_path
        volumes = {
            local_stores_path: {
                "bind": local_stores_path,
                "mode": "rw",
            }
        }
        orchestrator_run_id = str(uuid4())

        environment[ENV_ZENML_DOCKER_ORCHESTRATOR_RUN_ID] = orchestrator_run_id
        environment[ENV_ZENML_LOCAL_STORES_PATH] = local_stores_path

        image = self.get_image(snapshot=snapshot)
        image = settings.run_args.pop("image", image)

        user = None
        if sys.platform != "win32":
            user = os.getuid()
        user = settings.run_args.pop("user", user)

        run_args = copy.deepcopy(settings.run_args)
        docker_environment = run_args.pop("environment", {})
        docker_environment.update(environment)

        docker_volumes = run_args.pop("volumes", {})
        docker_volumes.update(volumes)

        extra_hosts = run_args.pop("extra_hosts", {})
        extra_hosts["host.docker.internal"] = "host-gateway"

        start_time = time.time()
        try:
            container = docker_client.containers.run(
                detach=True,
                image=image,
                entrypoint=command,
                command=arguments,
                user=user,
                volumes=docker_volumes,
                environment=docker_environment,
                extra_hosts=extra_hosts,
                **run_args,
            )
        except ContainerError as e:
            error_message = e.stderr.decode()
            raise RuntimeError(error_message)

        logger.info(
            "Pipeline run started in container `%s`.", container.short_id
        )

        if settings.synchronous:

            def _wait_for_run_to_finish() -> None:
                try:
                    for line in container.logs(stream=True):
                        logger.info(line.strip().decode())
                except ContainerError as e:
                    error_message = e.stderr.decode()
                    raise RuntimeError(error_message)

                run_duration = time.time() - start_time
                logger.info(
                    "Pipeline run has finished in `%s`.",
                    string_utils.get_human_readable_time(run_duration),
                )

            return SubmissionResult(
                wait_for_completion=_wait_for_run_to_finish,
            )
        else:
            return None

    @property
    def supported_execution_modes(self) -> List[ExecutionMode]:
        """Supported execution modes for this orchestrator.

        Returns:
            Supported execution modes for this orchestrator.
        """
        return [
            ExecutionMode.FAIL_FAST,
            ExecutionMode.STOP_ON_FAILURE,
            ExecutionMode.CONTINUE_ON_FAILURE,
        ]


class LocalDockerOrchestratorSettings(BaseSettings):
    """Local Docker orchestrator settings."""

    synchronous: bool = Field(
        default=True,
        description="Whether to run pipelines synchronously or asynchronously.",
    )
    run_args: Dict[str, Any] = Field(
        default={},
        description="Arguments to pass to the `docker run` call. "
        "(See https://docker-py.readthedocs.io/en/stable/containers.html for "
        "a list of what can be passed)",
    )


class LocalDockerOrchestratorConfig(
    BaseOrchestratorConfig, LocalDockerOrchestratorSettings
):
    """Local Docker orchestrator config."""

    @property
    def is_local(self) -> bool:
        """Checks if this stack component is running locally.

        Returns:
            True if this config is for a local component, False otherwise.
        """
        return True

    @property
    def is_synchronous(self) -> bool:
        """Whether the orchestrator runs synchronous or not.

        Returns:
            Whether the orchestrator runs synchronous or not.
        """
        return self.synchronous


class LocalDockerOrchestratorFlavor(BaseOrchestratorFlavor):
    """Flavor for the local Docker orchestrator."""

    @property
    def name(self) -> str:
        """Name of the orchestrator flavor.

        Returns:
            Name of the orchestrator flavor.
        """
        return "local_docker"

    @property
    def docs_url(self) -> Optional[str]:
        """A url to point at docs explaining this flavor.

        Returns:
            A flavor docs url.
        """
        return self.generate_default_docs_url()

    @property
    def sdk_docs_url(self) -> Optional[str]:
        """A url to point at SDK docs explaining this flavor.

        Returns:
            A flavor SDK docs url.
        """
        return self.generate_default_sdk_docs_url()

    @property
    def logo_url(self) -> str:
        """A url to represent the flavor in the dashboard.

        Returns:
            The flavor logo.
        """
        return "https://public-flavor-logos.s3.eu-central-1.amazonaws.com/orchestrator/docker.png"

    @property
    def config_class(self) -> Type[BaseOrchestratorConfig]:
        """Config class for the base orchestrator flavor.

        Returns:
            The config class.
        """
        return LocalDockerOrchestratorConfig

    @property
    def implementation_class(self) -> Type["LocalDockerOrchestrator"]:
        """Implementation class for this flavor.

        Returns:
            Implementation class for this flavor.
        """
        return LocalDockerOrchestrator



#  Copyright (c) ZenML GmbH 2025. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Run:AI integration constants and status mappings."""

from enum import Enum

from zenml.enums import ExecutionStatus

MAX_WORKLOAD_NAME_LENGTH = 50


class RunAIWorkloadStatus(str, Enum):
    """Run:AI workload status values."""

    PENDING = "pending"
    RUNNING = "running"
    SUCCEEDED = "succeeded"
    COMPLETED = "completed"
    SUCCESS = "success"
    FAILED = "failed"
    ERROR = "error"
    STOPPED = "stopped"
    STOPPING = "stopping"


RUNAI_STATUS_TO_EXECUTION_STATUS = {
    RunAIWorkloadStatus.PENDING: ExecutionStatus.INITIALIZING,
    RunAIWorkloadStatus.RUNNING: ExecutionStatus.RUNNING,
    RunAIWorkloadStatus.SUCCEEDED: ExecutionStatus.COMPLETED,
    RunAIWorkloadStatus.COMPLETED: ExecutionStatus.COMPLETED,
    RunAIWorkloadStatus.SUCCESS: ExecutionStatus.COMPLETED,
    RunAIWorkloadStatus.FAILED: ExecutionStatus.FAILED,
    RunAIWorkloadStatus.ERROR: ExecutionStatus.FAILED,
    RunAIWorkloadStatus.STOPPED: ExecutionStatus.FAILED,
    RunAIWorkloadStatus.STOPPING: ExecutionStatus.STOPPING,
}

# Frozen sets are used for quick status classification to avoid creating new sets on each call.
_TERMINAL_STATUSES = frozenset(
    {
        RunAIWorkloadStatus.SUCCEEDED,
        RunAIWorkloadStatus.COMPLETED,
        RunAIWorkloadStatus.SUCCESS,
        RunAIWorkloadStatus.FAILED,
        RunAIWorkloadStatus.ERROR,
        RunAIWorkloadStatus.STOPPED,
    }
)

_SUCCESS_STATUSES = frozenset(
    {
        RunAIWorkloadStatus.SUCCEEDED,
        RunAIWorkloadStatus.COMPLETED,
        RunAIWorkloadStatus.SUCCESS,
    }
)

_FAILURE_STATUSES = frozenset(
    {
        RunAIWorkloadStatus.FAILED,
        RunAIWorkloadStatus.ERROR,
        RunAIWorkloadStatus.STOPPED,
    }
)

_PENDING_STATUSES = frozenset({RunAIWorkloadStatus.PENDING})


def is_terminal_status(status: str) -> bool:
    """Check if a Run:AI status is terminal (workload finished).

    Args:
        status: The Run:AI workload status string.

    Returns:
        True if the status indicates the workload has finished.
    """
    try:
        return RunAIWorkloadStatus(status.lower()) in _TERMINAL_STATUSES
    except ValueError:
        return False


def is_success_status(status: str) -> bool:
    """Check if a Run:AI status indicates success.

    Args:
        status: The Run:AI workload status string.

    Returns:
        True if the status indicates successful completion.
    """
    try:
        return RunAIWorkloadStatus(status.lower()) in _SUCCESS_STATUSES
    except ValueError:
        return False


def is_failure_status(status: str) -> bool:
    """Check if a Run:AI status indicates failure.

    Args:
        status: The Run:AI workload status string.

    Returns:
        True if the status indicates failure.
    """
    try:
        return RunAIWorkloadStatus(status.lower()) in _FAILURE_STATUSES
    except ValueError:
        return False


def is_pending_status(status: str) -> bool:
    """Check if a Run:AI status indicates pending scheduling.

    Args:
        status: The Run:AI workload status string.

    Returns:
        True if the status indicates the workload is pending.
    """
    try:
        return RunAIWorkloadStatus(status.lower()) in _PENDING_STATUSES
    except ValueError:
        return False


def map_runai_status_to_execution_status(runai_status: str) -> ExecutionStatus:
    """Maps Run:AI workload status to ZenML ExecutionStatus.

    Args:
        runai_status: The Run:AI workload status string.

    Returns:
        The corresponding ZenML ExecutionStatus.
    """
    if not runai_status:
        return ExecutionStatus.RUNNING

    try:
        status_enum = RunAIWorkloadStatus(runai_status.lower())
        return RUNAI_STATUS_TO_EXECUTION_STATUS.get(
            status_enum, ExecutionStatus.RUNNING
        )
    except ValueError:
        return ExecutionStatus.RUNNING



#  Copyright (c) ZenML GmbH 2023. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Initialization for the ZenML MLflow integration.

The MLflow integrations currently enables you to use MLflow tracking as a
convenient way to visualize your experiment runs within the MLflow UI.
"""
from packaging import version
from typing import List, Type, Optional

from zenml.integrations.constants import MLFLOW
from zenml.integrations.integration import Integration
from zenml.stack import Flavor
import sys

from zenml.logger import get_logger

MLFLOW_MODEL_DEPLOYER_FLAVOR = "mlflow"
MLFLOW_MODEL_EXPERIMENT_TRACKER_FLAVOR = "mlflow"
MLFLOW_MODEL_REGISTRY_FLAVOR = "mlflow"

logger = get_logger(__name__)


class MlflowIntegration(Integration):
    """Definition of MLflow integration for ZenML."""

    NAME = MLFLOW

    REQUIREMENTS_IGNORED_ON_UNINSTALL = [
        "python-rapidjson",
        "pydantic",
        "numpy",
        "pandas",
    ]

    @classmethod
    def get_requirements(
        cls, target_os: Optional[str] = None, python_version: Optional[str] = None
    ) -> List[str]:
        """Method to get the requirements for the integration.

        Args:
            target_os: The target operating system to get the requirements for.
            python_version: The Python version to use for the requirements.

        Returns:
            A list of requirements.
        """
        from zenml.integrations.numpy import NumpyIntegration
        from zenml.integrations.pandas import PandasIntegration
        
        reqs = [
            "mlflow>=2.1.1,<4",
            # TODO: remove this requirement once rapidjson is fixed
            "python-rapidjson<1.15",
        ]

        reqs.extend(NumpyIntegration.get_requirements(target_os=target_os, python_version=python_version))
        reqs.extend(PandasIntegration.get_requirements(target_os=target_os, python_version=python_version))
        return reqs

    @classmethod
    def activate(cls) -> None:
        """Activate the MLflow integration."""
        from zenml.integrations.mlflow import services  # noqa

    @classmethod
    def flavors(cls) -> List[Type[Flavor]]:
        """Declare the stack component flavors for the MLflow integration.

        Returns:
            List of stack component flavors for this integration.
        """
        from zenml.integrations.mlflow.flavors import (
            MLFlowExperimentTrackerFlavor,
            MLFlowModelDeployerFlavor,
            MLFlowModelRegistryFlavor,
        )

        return [
            MLFlowModelDeployerFlavor,
            MLFlowExperimentTrackerFlavor,
            MLFlowModelRegistryFlavor,
        ]




#  Copyright (c) ZenML GmbH 2024. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Entrypoint of the Skypilot master/orchestrator VM."""

import argparse
import hashlib
import json
import shlex
import socket
import time
from typing import Any, Dict, Optional, cast

import sky

from zenml.client import Client
from zenml.entrypoints.step_entrypoint_configuration import (
    StepEntrypointConfiguration,
)
from zenml.enums import ExecutionStatus
from zenml.integrations.skypilot.flavors.skypilot_orchestrator_base_vm_config import (
    SkypilotBaseOrchestratorSettings,
)
from zenml.integrations.skypilot.orchestrators.skypilot_base_vm_orchestrator import (
    ENV_ZENML_SKYPILOT_ORCHESTRATOR_RUN_ID,
    SkypilotBaseOrchestrator,
)
from zenml.integrations.skypilot.utils import (
    create_docker_run_command,
    prepare_docker_setup,
    prepare_launch_kwargs,
    prepare_resources_kwargs,
    prepare_task_kwargs,
    sanitize_cluster_name,
    sky_job_get,
)
from zenml.logger import get_logger
from zenml.models import PipelineRunResponse
from zenml.orchestrators.dag_runner import NodeStatus, ThreadedDagRunner
from zenml.orchestrators.publish_utils import (
    publish_failed_pipeline_run,
)
from zenml.orchestrators.utils import get_config_environment_vars
from zenml.utils import env_utils

logger = get_logger(__name__)

_MAX_CLUSTER_NAME_LENGTH = 63
_CLUSTER_HASH_LENGTH = 16


def parse_args() -> argparse.Namespace:
    """Parse entrypoint arguments.

    Returns:
        Parsed args.
    """
    parser = argparse.ArgumentParser()
    parser.add_argument("--run_name", type=str, required=True)
    parser.add_argument("--snapshot_id", type=str, required=True)
    parser.add_argument("--run_id", type=str, required=False)
    return parser.parse_args()


def _normalize_for_json(value: Any) -> Any:
    """Converts nested values into deterministic JSON-serializable objects.

    Args:
        value: Value to normalize.

    Returns:
        A normalized representation safe for deterministic JSON serialization.
    """
    if isinstance(value, dict):
        return {
            str(key): _normalize_for_json(item)
            for key, item in sorted(
                value.items(), key=lambda entry: str(entry[0])
            )
        }
    if isinstance(value, (list, tuple)):
        return [_normalize_for_json(item) for item in value]
    if isinstance(value, set):
        normalized_items = [_normalize_for_json(item) for item in value]
        return sorted(
            normalized_items,
            key=lambda item: json.dumps(item, sort_keys=True),
        )
    return value


def _resource_config_digest(
    settings: SkypilotBaseOrchestratorSettings,
    default_instance_type: Optional[str],
) -> str:
    """Builds a stable digest for resource-affecting settings.

    Args:
        settings: Step-specific SkyPilot settings.
        default_instance_type: Default instance type of the orchestrator.

    Returns:
        A short deterministic hash for the resource configuration.
    """
    resource_config = {
        "instance_type": settings.instance_type or default_instance_type,
        "cpus": settings.cpus,
        "memory": settings.memory,
        "accelerators": settings.accelerators,
        "accelerator_args": settings.accelerator_args,
        "use_spot": settings.use_spot,
        "job_recovery": settings.job_recovery,
        "region": settings.region,
        "zone": settings.zone,
        "image_id": settings.image_id,
        "disk_size": settings.disk_size,
        "disk_tier": settings.disk_tier,
        "network_tier": settings.network_tier,
        "infra": settings.infra,
        "num_nodes": settings.num_nodes,
        "ports": settings.ports,
        "labels": settings.labels,
        "any_of": settings.any_of,
        "ordered": settings.ordered,
        "resources_settings": settings.resources_settings,
    }
    serialized = json.dumps(
        _normalize_for_json(resource_config),
        separators=(",", ":"),
        sort_keys=True,
    )
    return hashlib.sha256(serialized.encode("utf-8")).hexdigest()[
        :_CLUSTER_HASH_LENGTH
    ]


def _generate_cluster_name(
    orchestrator_run_id: str,
    settings: SkypilotBaseOrchestratorSettings,
    default_instance_type: Optional[str],
) -> str:
    """Generates a deterministic cluster name for a resource configuration.

    Args:
        orchestrator_run_id: Orchestrator run identifier.
        settings: Step-specific SkyPilot settings.
        default_instance_type: Default instance type of the orchestrator.

    Returns:
        Deterministic cluster name within the configured max length.
    """
    run_part = sanitize_cluster_name(orchestrator_run_id) or "run"
    digest = _resource_config_digest(settings, default_instance_type)

    max_run_part_len = (
        _MAX_CLUSTER_NAME_LENGTH - len("cluster--") - _CLUSTER_HASH_LENGTH
    )
    run_part = run_part[:max_run_part_len]

    return sanitize_cluster_name(f"cluster-{run_part}-{digest}")


def _resolve_pipeline_run(
    client: Client,
    run_name: str,
    snapshot_id: str,
    run_id: Optional[str],
) -> PipelineRunResponse:
    """Resolves the pipeline run for the orchestrator entrypoint.

    Args:
        client: Active ZenML client.
        run_name: Name passed to the entrypoint.
        snapshot_id: Snapshot ID of the run.
        run_id: Optional concrete run ID.

    Returns:
        The resolved pipeline run object.

    Raises:
        RuntimeError: If no matching run can be found.
    """
    if run_id:
        try:
            return client.get_pipeline_run(run_id)
        except Exception as e:
            raise RuntimeError(
                f"Unable to fetch pipeline run `{run_id}` passed via --run_id."
            ) from e

    runs = client.list_pipeline_runs(
        sort_by="desc:created",
        size=1,
        snapshot_id=snapshot_id,
        status=ExecutionStatus.INITIALIZING,
    )

    if getattr(runs, "total", None) == 0:
        raise RuntimeError(
            "Unable to resolve a pipeline run for snapshot "
            f"`{snapshot_id}` (run name `{run_name}`). Pass --run_id to "
            "the entrypoint to avoid race conditions."
        )

    return runs[0]


def main() -> None:
    """Entrypoint of the Skypilot master/orchestrator VM.

    This is the entrypoint of the Skypilot master/orchestrator VM. It is
    responsible for provisioning the VM and running the pipeline steps in
    separate VMs.

    The VM is provisioned using the `sky` library. The pipeline steps are run
    using the `sky` library as well.

    Raises:
        TypeError: If the active stack's orchestrator is not an instance of
            SkypilotBaseOrchestrator.
        ValueError: If the active stack's container registry is None.
        Exception: If the orchestration or one of the steps fails.
    """
    logger.info("Skypilot orchestrator VM started.")

    args = parse_args()
    orchestrator_run_id = socket.gethostname()

    run = None

    try:
        client = Client()
        snapshot = client.get_snapshot(args.snapshot_id)

        pipeline_dag = {
            step_name: step.spec.upstream_steps
            for step_name, step in snapshot.step_configurations.items()
        }
        step_command = StepEntrypointConfiguration.get_entrypoint_command()
        entrypoint_str = " ".join(shlex.quote(token) for token in step_command)

        active_stack = client.active_stack

        orchestrator = active_stack.orchestrator
        if not isinstance(orchestrator, SkypilotBaseOrchestrator):
            raise TypeError(
                "The active stack's orchestrator is not an instance of SkypilotBaseOrchestrator."
            )

        orchestrator.setup_credentials()

        orchestrator.prepare_environment_variable(set=True)

        container_registry = active_stack.container_registry
        if container_registry is None:
            raise ValueError("Container registry cannot be None.")

        setup, task_envs = prepare_docker_setup(
            container_registry_uri=container_registry.config.uri,
            credentials=container_registry.credentials,
            use_sudo=False,  # Entrypoint doesn't use sudo
        )

        unique_resource_configs: Dict[str, str] = {}
        for step_name, step in snapshot.step_configurations.items():
            settings = cast(
                SkypilotBaseOrchestratorSettings,
                orchestrator.get_settings(step),
            )
            cluster_name = _generate_cluster_name(
                orchestrator_run_id=orchestrator_run_id,
                settings=settings,
                default_instance_type=orchestrator.DEFAULT_INSTANCE_TYPE,
            )
            unique_resource_configs[step_name] = cluster_name

        run = _resolve_pipeline_run(
            client=client,
            run_name=args.run_name,
            snapshot_id=args.snapshot_id,
            run_id=args.run_id,
        )

        logger.info("Fetching pipeline run: %s", run.id)

        shared_env, secrets = get_config_environment_vars()
        shared_env[ENV_ZENML_SKYPILOT_ORCHESTRATOR_RUN_ID] = (
            orchestrator_run_id
        )

        def run_step_on_skypilot_vm(step_name: str) -> None:
            """Run a pipeline step in a separate Skypilot VM.

            Args:
                step_name: Name of the step.

            Raises:
                Exception: If the step execution fails.
            """
            logger.info(f"Running step `{step_name}` on a VM...")
            try:
                cluster_name = unique_resource_configs[step_name]

                image = SkypilotBaseOrchestrator.get_image(
                    snapshot=snapshot, step_name=step_name
                )

                step_args = (
                    StepEntrypointConfiguration.get_entrypoint_arguments(
                        step_name=step_name, snapshot_id=snapshot.id
                    )
                )
                arguments_str = " ".join(
                    shlex.quote(token) for token in step_args
                )

                step = snapshot.step_configurations[step_name]
                settings = cast(
                    SkypilotBaseOrchestratorSettings,
                    orchestrator.get_settings(step),
                )
                step_env = shared_env.copy()
                step_env.update(
                    env_utils.get_runtime_environment(
                        config=step.config, stack=active_stack
                    )
                )
                step_env.update(secrets)

                run_command = create_docker_run_command(
                    image=image,
                    entrypoint_str=entrypoint_str,
                    arguments_str=arguments_str,
                    environment=step_env,
                    docker_run_args=settings.docker_run_args,
                    use_sudo=False,  # Entrypoint doesn't use sudo
                )

                task_name = f"{snapshot.id}-{step_name}-{time.time()}"

                step_task_envs = step_env.copy()
                step_task_envs.update(task_envs)
                task_kwargs = prepare_task_kwargs(
                    settings=settings,
                    run_command=run_command,
                    setup=setup,
                    task_envs=step_task_envs,
                    task_name=task_name,
                )

                task = sky.Task(**task_kwargs)

                resources_kwargs = prepare_resources_kwargs(
                    cloud=orchestrator.cloud,
                    settings=settings,
                    default_instance_type=orchestrator.DEFAULT_INSTANCE_TYPE,
                )

                task = task.set_resources(sky.Resources(**resources_kwargs))

                launch_kwargs = prepare_launch_kwargs(settings=settings)

                # sky.launch now returns a request ID (async). Capture it so we can
                # optionally stream logs and block until completion when desired.
                launch_request_id = sky.launch(
                    task,
                    cluster_name,
                    **launch_kwargs,
                )
                sky_job_get(
                    launch_request_id, settings.stream_logs, cluster_name
                )

                # Pop the resource configuration for this step
                unique_resource_configs.pop(step_name)

                if cluster_name in unique_resource_configs.values():
                    logger.info(
                        f"Resource configuration for cluster '{cluster_name}' "
                        "is used by subsequent steps. Skipping the deprovisioning of "
                        "the cluster."
                    )
                else:
                    logger.info(
                        f"Resource configuration for cluster '{cluster_name}' "
                        "is not used by subsequent steps. deprovisioning the cluster."
                    )
                    try:
                        down_request_id = sky.down(cluster_name)
                        sky.stream_and_get(down_request_id)
                    except Exception as cleanup_error:
                        logger.error(
                            "Failed to deprovision cluster "
                            f"'{cluster_name}': {cleanup_error}. "
                            f"Resources may still be running. Run `sky down "
                            f"{cluster_name}` manually to clean up."
                        )

                logger.info(
                    f"Running step `{step_name}` on a VM is completed."
                )

            except Exception as e:
                logger.error(f"Failed while launching step `{step_name}`: {e}")
                raise

        dag_runner = ThreadedDagRunner(
            dag=pipeline_dag, run_fn=run_step_on_skypilot_vm
        )
        dag_runner.run()

        failed_nodes = []
        for node in dag_runner.nodes:
            if dag_runner.node_states[node] == NodeStatus.FAILED:
                failed_nodes.append(node)

        if failed_nodes:
            raise Exception(f"One or more steps failed: {failed_nodes}")

    except Exception as e:
        logger.error(f"Orchestrator failed: {e}")

        # Try to mark the pipeline run as failed
        if run:
            publish_failed_pipeline_run(run.id)
            logger.info("Marked pipeline run as failed in ZenML.")
        raise


if __name__ == "__main__":
    main()



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Implementation of the PyTorch DataLoader materializer."""

import os
from typing import Any, ClassVar, Type

import cloudpickle

from zenml.io import fileio
from zenml.materializers.base_materializer import BaseMaterializer

DEFAULT_FILENAME = "obj.pt"


class BasePyTorchMaterializer(BaseMaterializer):
    """Base class for PyTorch materializers."""

    FILENAME: ClassVar[str] = DEFAULT_FILENAME
    SKIP_REGISTRATION: ClassVar[bool] = True

    def load(self, data_type: Type[Any]) -> Any:
        """Uses `torch.load` to load a PyTorch object.

        Args:
            data_type: The type of the object to load.

        Returns:
            The loaded PyTorch object.
        """
        import torch

        with fileio.open(os.path.join(self.uri, self.FILENAME), "rb") as f:
            # NOTE (security): The `torch.load` function uses `pickle` as
            # the default unpickler, which is NOT secure. This materializer
            # is intended for use with trusted data sources.
            return torch.load(f, weights_only=False)  # nosec

    def save(self, obj: Any) -> None:
        """Uses `torch.save` to save a PyTorch object.

        Args:
            obj: The PyTorch object to save.
        """
        import torch

        with fileio.open(os.path.join(self.uri, self.FILENAME), "wb") as f:
            # NOTE (security): The `torch.save` function uses `cloudpickle` as
            # the default unpickler, which is NOT secure. This materializer
            # is intended for use with trusted data sources.
            torch.save(obj, f, pickle_module=cloudpickle)  # nosec


# Alias for the BasePyTorchMaterializer class, allowing users that have already used
# the old name to continue using it without breaking their code.
# 'BasePyTorchMaterializer' or 'BasePyTorchMaterliazer' to refer to the same class.
BasePyTorchMaterliazer = BasePyTorchMaterializer



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""WhyLabs whylogs data validator flavor."""

from typing import TYPE_CHECKING, Optional, Type

from pydantic import Field

from zenml.config.base_settings import BaseSettings
from zenml.data_validators.base_data_validator import (
    BaseDataValidatorConfig,
    BaseDataValidatorFlavor,
)
from zenml.integrations.whylogs import WHYLOGS_DATA_VALIDATOR_FLAVOR
from zenml.stack.authentication_mixin import AuthenticationConfigMixin

if TYPE_CHECKING:
    from zenml.integrations.whylogs.data_validators import WhylogsDataValidator


class WhylogsDataValidatorSettings(BaseSettings):
    """Settings for the Whylogs data validator."""

    enable_whylabs: bool = Field(
        False,
        description="If set to `True` for a step, all the whylogs data "
        "profile views returned by the step will automatically be uploaded "
        "to the Whylabs platform if Whylabs credentials are configured.",
    )
    dataset_id: Optional[str] = Field(
        None,
        description="Dataset ID to use when uploading profiles to Whylabs.",
    )


class WhylogsDataValidatorConfig(
    BaseDataValidatorConfig,
    AuthenticationConfigMixin,
    WhylogsDataValidatorSettings,
):
    """Config for the whylogs data validator."""


class WhylogsDataValidatorFlavor(BaseDataValidatorFlavor):
    """Whylogs data validator flavor."""

    @property
    def name(self) -> str:
        """Name of the flavor.

        Returns:
            The name of the flavor.
        """
        return WHYLOGS_DATA_VALIDATOR_FLAVOR

    @property
    def docs_url(self) -> Optional[str]:
        """A url to point at docs explaining this flavor.

        Returns:
            A flavor docs url.
        """
        return self.generate_default_docs_url()

    @property
    def sdk_docs_url(self) -> Optional[str]:
        """A url to point at SDK docs explaining this flavor.

        Returns:
            A flavor SDK docs url.
        """
        return self.generate_default_sdk_docs_url()

    @property
    def logo_url(self) -> str:
        """A url to represent the flavor in the dashboard.

        Returns:
            The flavor logo.
        """
        return "https://public-flavor-logos.s3.eu-central-1.amazonaws.com/data_validator/whylogs.png"

    @property
    def config_class(self) -> Type[WhylogsDataValidatorConfig]:
        """Returns `WhylogsDataValidatorConfig` config class.

        Returns:
                The config class.
        """
        return WhylogsDataValidatorConfig

    @property
    def implementation_class(self) -> Type["WhylogsDataValidator"]:
        """Implementation class for this flavor.

        Returns:
            The implementation class.
        """
        from zenml.integrations.whylogs.data_validators import (
            WhylogsDataValidator,
        )

        return WhylogsDataValidator



#  Copyright (c) ZenML GmbH 2024. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Pigeon annotator.

Credit for the implementation of this code to @agermanidis in the
Pigeon package and library. This code has been slightly modified to
fit the ZenML framework. We use the modified code directly here because
the original package (and code) is no longer compatible with more recent
versions of ipywidgets.

https://github.com/agermanidis/pigeon
"""

import json
import os
from datetime import datetime
from typing import Any, List, Optional, Tuple, cast

import ipywidgets as widgets  # type: ignore
from IPython.core.display_functions import clear_output, display

from zenml.annotators.base_annotator import BaseAnnotator
from zenml.integrations.pigeon.flavors.pigeon_annotator_flavor import (
    PigeonAnnotatorConfig,
)
from zenml.logger import get_logger

logger = get_logger(__name__)


class PigeonAnnotator(BaseAnnotator):
    """Annotator for using Pigeon in Jupyter notebooks."""

    @property
    def config(self) -> PigeonAnnotatorConfig:
        """Get the Pigeon annotator config.

        Returns:
            The Pigeon annotator config.
        """
        return cast(PigeonAnnotatorConfig, self._config)

    def get_url(self) -> str:
        """Get the URL of the Pigeon annotator.

        Raises:
            NotImplementedError: Pigeon annotator does not have a URL.
        """
        raise NotImplementedError("Pigeon annotator does not have a URL.")

    def get_url_for_dataset(self, dataset_name: str) -> str:
        """Get the URL of the Pigeon annotator for a specific dataset.

        Args:
            dataset_name: Name of the dataset (annotation file).

        Raises:
            NotImplementedError: Pigeon annotator does not have a URL.
        """
        raise NotImplementedError("Pigeon annotator does not have a URL.")

    def get_datasets(self) -> List[str]:
        """Get a list of datasets (annotation files) in the output directory.

        Returns:
            A list of dataset names (annotation file names) (or empty list when no datasets are present).
        """
        output_dir = self.config.output_dir
        try:
            return [f for f in os.listdir(output_dir) if f.endswith(".txt")]
        except FileNotFoundError:
            return []

    def get_dataset_names(self) -> List[str]:
        """List dataset names (annotation file names) in the output directory.

        Returns:
            A list of dataset names (annotation file names).
        """
        return self.get_datasets()

    def get_dataset_stats(self, dataset_name: str) -> Tuple[int, int]:
        """List labeled and unlabeled examples in a dataset (annotation file).

        Args:
            dataset_name: Name of the dataset (annotation file).

        Returns:
            A tuple containing (num_labeled_examples, num_unlabeled_examples).
        """
        dataset_path = os.path.join(self.config.output_dir, dataset_name)
        num_labeled_examples = 0
        # Placeholder as logic to determine this is not implemented
        num_unlabeled_examples = 0

        try:
            with open(dataset_path, "r") as file:
                num_labeled_examples = sum(1 for _ in file)
        except FileNotFoundError:
            logger.error(f"File not found: {dataset_path}")

        return num_labeled_examples, num_unlabeled_examples

    def _annotate(
        self,
        data: List[Any],
        options: List[str],
        display_fn: Optional[Any] = None,
    ) -> List[Tuple[Any, Any]]:
        """Internal method to build an interactive widget for annotating.

        Args:
            data: List of examples to annotate.
            options: List of labels to choose from.
            display_fn: Optional function to display examples.

        Returns:
            A list of tuples containing (example, label) for each annotated example.
        """
        examples = list(data)
        annotations = []
        current_index = 0
        out = widgets.Output()

        def show_next() -> None:
            nonlocal current_index
            if current_index >= len(examples):
                with out:
                    clear_output(wait=True)
                    logger.info("Annotation done.")
                return
            with out:
                clear_output(wait=True)
                if display_fn:
                    display_fn(examples[current_index])
                else:
                    display(examples[current_index])

        def add_annotation(btn: widgets.Button) -> None:
            """Add an annotation to the list of annotations.

            Args:
                btn: The button that triggered the event.
            """
            nonlocal current_index
            annotation = btn.description
            annotations.append((examples[current_index], annotation))
            current_index += 1
            show_next()

        def submit_annotations(btn: widgets.Button) -> None:
            """Submit all annotations and save them to a file.

            Args:
                btn: The button that triggered the event.
            """
            self._save_annotations(annotations)
            with out:
                clear_output(wait=True)
                logger.info("Annotations saved.")

        count_label = widgets.Label()
        display(count_label)

        buttons = []
        for label in options:
            btn = widgets.Button(description=label)
            btn.on_click(add_annotation)
            buttons.append(btn)

        submit_btn = widgets.Button(
            description="Save labels", button_style="success"
        )
        submit_btn.on_click(submit_annotations)
        buttons.append(submit_btn)

        navigation_box = widgets.HBox(buttons)
        display(navigation_box)
        display(out)
        show_next()

        return annotations

    def launch(self, **kwargs: Any) -> None:
        """Launch the Pigeon annotator in the Jupyter notebook.

        Args:
            **kwargs: Additional keyword arguments to pass to the annotation client.

        Raises:
            NotImplementedError: Pigeon annotator does not support launching with a URL.
        """
        raise NotImplementedError(
            "Pigeon annotator does not support launching with a URL."
        )

    def annotate(
        self,
        data: List[Any],
        options: List[str],
        display_fn: Optional[Any] = None,
    ) -> List[Tuple[Any, Any]]:
        """Annotate with the Pigeon annotator in the Jupyter notebook.

        Args:
            data: List of examples to annotate.
            options: List of labels to choose from.
            display_fn: Optional function to display examples.

        Returns:
            A list of tuples containing (example, label) for each annotated example.
        """
        annotations = self._annotate(data, options, display_fn)
        return annotations

    def _save_annotations(self, annotations: List[Tuple[Any, Any]]) -> None:
        """Save annotations to a file with a unique date-time suffix.

        Args:
            annotations: List of tuples containing (example, label) for each annotated example.
        """
        output_dir = self.config.output_dir
        os.makedirs(output_dir, exist_ok=True)
        timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
        output_file = os.path.join(output_dir, f"annotations_{timestamp}.json")
        with open(output_file, "w") as f:
            json.dump(annotations, f)

    def add_dataset(self, **kwargs: Any) -> Any:
        """Add a dataset (annotation file) to the Pigeon annotator.

        Args:
            **kwargs: keyword arguments.

        Raises:
            NotImplementedError: Pigeon annotator does not support adding datasets.
        """
        raise NotImplementedError(
            "Pigeon annotator does not support adding datasets."
        )

    def delete_dataset(self, **kwargs: Any) -> None:
        """Delete a dataset (annotation file).

        Takes the `dataset_name` argument from the kwargs.

        Args:
            **kwargs: Keyword arguments containing the `dataset_name` to delete.

        Raises:
            ValueError: Dataset name is required to delete a dataset.
        """
        dataset_name = kwargs.get("dataset_name")
        if not dataset_name:
            raise ValueError(
                "Dataset name (`dataset_name`) is required to delete a dataset."
            )
        dataset_path = os.path.join(self.config.output_dir, dataset_name)
        os.remove(dataset_path)

    def get_dataset(self, **kwargs: Any) -> List[Tuple[Any, Any]]:
        """Get the annotated examples from a dataset (annotation file).

        Takes the `dataset_name` argument from the kwargs.

        Args:
            **kwargs: Keyword arguments containing the `dataset_name` to retrieve.

        Returns:
            A list of tuples containing (example, label) for each annotated
            example.

        Raises:
            ValueError: Dataset name is required to retrieve a dataset.
        """
        dataset_name = kwargs.get("dataset_name")
        if not dataset_name:
            raise ValueError(
                "Dataset name (`dataset_name`) is required to retrieve a dataset."
            )
        dataset_path = os.path.join(self.config.output_dir, dataset_name)
        with open(dataset_path, "r") as f:
            annotations = json.load(f)
        return cast(List[Tuple[Any, Any]], annotations)

    def get_labeled_data(self, **kwargs: Any) -> List[Tuple[Any, Any]]:
        """Get the labeled examples from a dataset (annotation file).

        Takes the `dataset_name` argument from the kwargs.

        Args:
            **kwargs: Keyword arguments containing the `dataset_name` to retrieve.

        Returns:
            A list of tuples containing (example, label) for each labeled
            example.

        Raises:
            ValueError: Dataset name is required to retrieve labeled data.
        """
        if dataset_name := kwargs.get("dataset_name"):
            return self.get_dataset(dataset_name=dataset_name)
        else:
            raise ValueError(
                "Dataset name (`dataset_name`) is required to retrieve labeled data."
            )

    def get_unlabeled_data(self, **kwargs: Any) -> Any:
        """Get the unlabeled examples from a dataset (annotation file).

        Args:
            **kwargs: keyword arguments.

        Raises:
            NotImplementedError: Pigeon annotator does not support retrieving unlabeled data.
        """
        raise NotImplementedError(
            "Pigeon annotator does not support retrieving unlabeled data."
        )



#  Copyright (c) ZenML GmbH 2025. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Base class for all ZenML deployers."""

from zenml.exceptions import EntityExistsError
from zenml.logger import get_logger

logger = get_logger(__name__)


class DeployerError(Exception):
    """Base class for deployer errors."""


class DeploymentAlreadyExistsError(EntityExistsError, DeployerError):
    """Error raised when a deployment already exists."""


class DeploymentNotFoundError(KeyError, DeployerError):
    """Error raised when a deployment is not found."""


class DeploymentProvisionError(DeployerError):
    """Error raised when a deployment provisioning fails."""


class DeploymentTimeoutError(DeployerError):
    """Error raised when a deployment provisioning or deprovisioning times out."""


class DeploymentDeprovisionError(DeployerError):
    """Error raised when a deployment deprovisioning fails."""


class DeploymentLogsNotFoundError(KeyError, DeployerError):
    """Error raised when pipeline logs are not found."""


class DeploymentDeployerMismatchError(DeployerError):
    """Error raised when a deployment is not managed by this deployer."""


class DeploymentSnapshotMismatchError(DeployerError):
    """Error raised when a deployment snapshot does not match the current deployer."""


class DeploymentHTTPError(DeployerError):
    """Error raised when an HTTP request to a deployment fails."""


class DeploymentInvalidParametersError(DeployerError):
    """Error raised when the parameters for a deployment are invalid."""



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""GCP Cloud Run deployer flavor."""

from typing import TYPE_CHECKING, Dict, Optional, Type

from pydantic import Field

from zenml.deployers.base_deployer import (
    BaseDeployerConfig,
    BaseDeployerFlavor,
    BaseDeployerSettings,
)
from zenml.integrations.gcp import (
    GCP_DEPLOYER_FLAVOR,
    GCP_RESOURCE_TYPE,
)
from zenml.integrations.gcp.google_credentials_mixin import (
    GoogleCredentialsConfigMixin,
)
from zenml.models import ServiceConnectorRequirements

if TYPE_CHECKING:
    from zenml.integrations.gcp.deployers import GCPDeployer


class GCPDeployerSettings(BaseDeployerSettings):
    """Settings for the GCP Cloud Run deployer."""

    location: str = Field(
        default="europe-west3",
        description="Name of GCP region where the pipeline will be deployed. "
        "Cloud Run is available in specific regions: "
        "https://cloud.google.com/run/docs/locations",
    )
    service_name_prefix: str = Field(
        default="zenml-",
        description="Prefix for service names in Cloud Run to avoid naming "
        "conflicts.",
    )

    # Timeout and execution configuration
    timeout_seconds: int = Field(
        default=300,
        ge=1,
        le=3600,
        description="Request timeout in seconds. Maximum is 3600 (1 hour).",
    )

    # Networking and security
    ingress: str = Field(
        default="all",
        description="Ingress settings for the service. "
        "Options: 'all', 'internal', 'internal-and-cloud-load-balancing'.",
    )

    vpc_connector: Optional[str] = Field(
        default=None,
        description="VPC connector for private networking. "
        "Format: projects/PROJECT_ID/locations/LOCATION/connectors/CONNECTOR_NAME",
    )

    # Service account and IAM
    service_account: Optional[str] = Field(
        default=None,
        description="Service account email to run the Cloud Run service. "
        "If not specified, uses the default Compute Engine service account.",
    )

    # Environment and configuration
    environment_variables: Dict[str, str] = Field(
        default_factory=dict,
        description="Environment variables to set in the Cloud Run service.",
    )

    # Labels and annotations
    labels: Dict[str, str] = Field(
        default_factory=dict,
        description="Labels to apply to the Cloud Run service.",
    )

    annotations: Dict[str, str] = Field(
        default_factory=dict,
        description="Annotations to apply to the Cloud Run service.",
    )

    # Cloud Run specific settings
    execution_environment: str = Field(
        default="gen2",
        description="Execution environment generation. Options: 'gen1', 'gen2'.",
    )

    # Deployment configuration
    traffic_allocation: Dict[str, int] = Field(
        default_factory=lambda: {"LATEST": 100},
        description="Traffic allocation between revisions. "
        "Keys are revision names or 'LATEST', values are percentages.",
    )

    allow_unauthenticated: bool = Field(
        default=True,
        description="Whether to allow unauthenticated requests to the service.",
    )

    # Secret management configuration
    use_secret_manager: bool = Field(
        default=True,
        description="Whether to store sensitive environment variables in GCP "
        "Secret Manager instead of directly in the Cloud Run service "
        "configuration.",
    )

    secret_name_prefix: str = Field(
        default="zenml-",
        description="Prefix for secret names in Secret Manager to avoid naming "
        "conflicts.",
    )


class GCPDeployerConfig(
    BaseDeployerConfig,
    GoogleCredentialsConfigMixin,
    GCPDeployerSettings,
):
    """Configuration for the GCP Cloud Run deployer."""

    @property
    def is_remote(self) -> bool:
        """Checks if this stack component is running remotely.

        This designation is used to determine if the stack component can be
        used with a local ZenML database or if it requires a remote ZenML
        server.

        Returns:
            True if this config is for a remote component, False otherwise.
        """
        return True


class GCPDeployerFlavor(BaseDeployerFlavor):
    """GCP Cloud Run deployer flavor."""

    @property
    def name(self) -> str:
        """Name of the orchestrator flavor.

        Returns:
            Name of the orchestrator flavor.
        """
        return GCP_DEPLOYER_FLAVOR

    @property
    def display_name(self) -> str:
        """Display name of the flavor.

        Returns:
            The display name of the flavor.
        """
        return "GCP"

    @property
    def service_connector_requirements(
        self,
    ) -> Optional[ServiceConnectorRequirements]:
        """Service connector resource requirements for service connectors.

        Specifies resource requirements that are used to filter the available
        service connector types that are compatible with this flavor.

        Returns:
            Requirements for compatible service connectors, if a service
            connector is required for this flavor.
        """
        return ServiceConnectorRequirements(
            resource_type=GCP_RESOURCE_TYPE,
        )

    @property
    def docs_url(self) -> Optional[str]:
        """A url to point at docs explaining this flavor.

        Returns:
            A flavor docs url.
        """
        return self.generate_default_docs_url()

    @property
    def sdk_docs_url(self) -> Optional[str]:
        """A url to point at SDK docs explaining this flavor.

        Returns:
            A flavor SDK docs url.
        """
        return self.generate_default_sdk_docs_url()

    @property
    def logo_url(self) -> str:
        """A url to represent the flavor in the dashboard.

        Returns:
            The flavor logo.
        """
        return "https://public-flavor-logos.s3.eu-central-1.amazonaws.com/deployer/google-cloud-run.svg"

    @property
    def config_class(self) -> Type[GCPDeployerConfig]:
        """Returns the GCPDeployerConfig config class.

        Returns:
                The config class.
        """
        return GCPDeployerConfig

    @property
    def implementation_class(self) -> Type["GCPDeployer"]:
        """Implementation class for this flavor.

        Returns:
            Implementation class for this flavor.
        """
        from zenml.integrations.gcp.deployers import GCPDeployer

        return GCPDeployer



"""remove logs unique constraint [8e36b8227160].

Revision ID: 8e36b8227160
Revises: 73c397e3c3ba
Create Date: 2026-01-15 17:31:19.491562

"""

import sqlalchemy as sa
import sqlmodel
from alembic import op

# revision identifiers, used by Alembic.
revision = "8e36b8227160"
down_revision = "73c397e3c3ba"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    with op.batch_alter_table("logs", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column(
                "project_id", sqlmodel.sql.sqltypes.GUID(), nullable=True
            )
        )
        batch_op.add_column(
            sa.Column("user_id", sqlmodel.sql.sqltypes.GUID(), nullable=True)
        )
        batch_op.add_column(
            sa.Column(
                "log_key", sqlmodel.sql.sqltypes.AutoString(), nullable=True
            )
        )
        batch_op.create_unique_constraint("unique_log_key", ["log_key"])
        batch_op.drop_constraint(
            "unique_source_per_run_and_step", type_="unique"
        )

    connection = op.get_bind()
    logs = sa.table(
        "logs",
        sa.column("project_id"),
        sa.column("user_id"),
        sa.column("step_run_id"),
        sa.column("pipeline_run_id"),
    )
    step_run = sa.table(
        "step_run",
        sa.column("id"),
        sa.column("project_id"),
        sa.column("user_id"),
    )
    pipeline_run = sa.table(
        "pipeline_run",
        sa.column("id"),
        sa.column("project_id"),
        sa.column("user_id"),
    )

    connection.execute(
        sa.update(logs)
        .where(logs.c.step_run_id.is_not(None))
        .values(
            project_id=sa.select(step_run.c.project_id)
            .where(step_run.c.id == logs.c.step_run_id)
            .scalar_subquery(),
            user_id=sa.select(step_run.c.user_id)
            .where(step_run.c.id == logs.c.step_run_id)
            .scalar_subquery(),
        )
    )

    connection.execute(
        sa.update(logs)
        .where(logs.c.step_run_id.is_(None))
        .where(logs.c.pipeline_run_id.is_not(None))
        .values(
            project_id=sa.select(pipeline_run.c.project_id)
            .where(pipeline_run.c.id == logs.c.pipeline_run_id)
            .scalar_subquery(),
            user_id=sa.select(pipeline_run.c.user_id)
            .where(pipeline_run.c.id == logs.c.pipeline_run_id)
            .scalar_subquery(),
        )
    )

    connection.execute(sa.delete(logs).where(logs.c.project_id.is_(None)))

    with op.batch_alter_table("logs", schema=None) as batch_op:
        batch_op.alter_column(
            "project_id",
            existing_type=sqlmodel.sql.sqltypes.GUID(),
            nullable=False,
        )
        batch_op.create_foreign_key(
            "fk_logs_project_id_project",
            "project",
            ["project_id"],
            ["id"],
            ondelete="CASCADE",
        )
        batch_op.create_foreign_key(
            "fk_logs_user_id_user",
            "user",
            ["user_id"],
            ["id"],
            ondelete="SET NULL",
        )


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    with op.batch_alter_table("logs", schema=None) as batch_op:
        batch_op.drop_constraint("fk_logs_user_id_user", type_="foreignkey")
        batch_op.drop_constraint(
            "fk_logs_project_id_project", type_="foreignkey"
        )
        batch_op.drop_constraint("unique_log_key", type_="unique")
        batch_op.drop_column("log_key")
        batch_op.drop_column("user_id")
        batch_op.drop_column("project_id")
        batch_op.create_unique_constraint(
            "unique_source_per_run_and_step",
            ["source", "pipeline_run_id", "step_run_id"],
        )



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Entrypoint of the Kubernetes master/orchestrator pod."""

import argparse
import random
import socket
import threading
import time
from contextlib import nullcontext
from typing import Any, ContextManager, Dict, List, Optional, Tuple, cast
from uuid import UUID

from kubernetes import client as k8s_client
from kubernetes.client.rest import ApiException

from zenml.client import Client
from zenml.entrypoints.step_entrypoint_configuration import (
    StepEntrypointConfiguration,
)
from zenml.enums import ExecutionMode, ExecutionStatus, MetadataResourceTypes
from zenml.exceptions import AuthorizationException
from zenml.integrations.kubernetes import kube_utils
from zenml.integrations.kubernetes.constants import (
    ENV_ZENML_KUBERNETES_RUN_ID,
    KUBERNETES_SECRET_TOKEN_KEY_NAME,
    ORCHESTRATOR_RUN_ID_ANNOTATION_KEY,
    RUN_ID_ANNOTATION_KEY,
    STEP_NAME_ANNOTATION_KEY,
)
from zenml.integrations.kubernetes.flavors.kubernetes_orchestrator_flavor import (
    KubernetesOrchestratorSettings,
)
from zenml.integrations.kubernetes.manifest_utils import (
    build_job_manifest,
    build_pod_manifest,
    pod_template_manifest_from_pod,
)
from zenml.integrations.kubernetes.orchestrators.dag_runner import (
    DagRunner,
    InterruptMode,
    Node,
    NodeStatus,
)
from zenml.integrations.kubernetes.orchestrators.kubernetes_orchestrator import (
    KubernetesOrchestrator,
)
from zenml.logger import get_logger
from zenml.models import (
    PipelineRunResponse,
    PipelineRunUpdate,
    PipelineSnapshotResponse,
    RunMetadataResource,
    StepRunResponse,
)
from zenml.orchestrators import publish_utils
from zenml.orchestrators.step_run_utils import (
    StepRunRequestFactory,
    fetch_step_runs_by_names,
    publish_cached_step_run,
)
from zenml.orchestrators.utils import (
    get_config_environment_vars,
)
from zenml.pipelines.run_utils import create_placeholder_run
from zenml.utils import env_utils
from zenml.utils.logging_utils import (
    LoggingContext,
    is_pipeline_logging_enabled,
    setup_logging_context,
)
from zenml.utils.time_utils import utc_now

logger = get_logger(__name__)


def parse_args() -> argparse.Namespace:
    """Parse entrypoint arguments.

    Returns:
        Parsed args.
    """
    parser = argparse.ArgumentParser()
    parser.add_argument("--snapshot_id", type=str, required=True)
    parser.add_argument("--run_id", type=str, required=False)
    return parser.parse_args()


def _get_orchestrator_job_state(
    batch_api: k8s_client.BatchV1Api,
    namespace: str,
    job_name: str,
    api_request_timeout: Optional[int],
) -> Tuple[Optional[UUID], Optional[str]]:
    """Get the existing status of the orchestrator job.

    Args:
        batch_api: The batch api.
        namespace: The namespace.
        job_name: The name of the orchestrator job.
        api_request_timeout: The request timeout in seconds.

    Returns:
        The run id and orchestrator run id.
    """
    run_id = None
    orchestrator_run_id = None

    job = kube_utils.get_job(
        batch_api=batch_api,
        namespace=namespace,
        job_name=job_name,
        api_request_timeout=api_request_timeout,
    )

    if job.metadata and job.metadata.annotations:
        annotations = job.metadata.annotations

        run_id = annotations.get(RUN_ID_ANNOTATION_KEY, None)
        orchestrator_run_id = annotations.get(
            ORCHESTRATOR_RUN_ID_ANNOTATION_KEY, None
        )

    return UUID(run_id) if run_id else None, orchestrator_run_id


def _reconstruct_nodes(
    snapshot: PipelineSnapshotResponse,
    pipeline_run: PipelineRunResponse,
    namespace: str,
    batch_api: k8s_client.BatchV1Api,
    api_request_timeout: Optional[int] = None,
) -> List[Node]:
    """Reconstruct the nodes from the pipeline run.

    Args:
        snapshot: The snapshot.
        pipeline_run: The pipeline run.
        namespace: The namespace.
        batch_api: The batch api.
        api_request_timeout: The request timeout in seconds.

    Returns:
        The reconstructed nodes.
    """
    nodes = {
        step_name: Node(id=step_name, upstream_nodes=step.spec.upstream_steps)
        for step_name, step in snapshot.step_configurations.items()
    }

    for step_name, existing_step_run in pipeline_run.steps.items():
        node = nodes[step_name]
        if existing_step_run.status.is_successful:
            node.status = NodeStatus.COMPLETED
        elif existing_step_run.status.is_finished:
            node.status = NodeStatus.FAILED

    job_list = kube_utils.list_jobs(
        batch_api=batch_api,
        namespace=namespace,
        label_selector=f"run_id={pipeline_run.id}",
        api_request_timeout=api_request_timeout,
    )
    for job in job_list.items:
        annotations = job.metadata.annotations or {}
        if step_name_annotation := annotations.get(
            STEP_NAME_ANNOTATION_KEY, None
        ):
            node = nodes[str(step_name_annotation)]
            node.metadata["job_name"] = job.metadata.name

            if node.status == NodeStatus.NOT_READY:
                # The step is not finished in the ZenML database, so we base it
                # on the job status.
                node_status = NodeStatus.RUNNING
                if job.status.conditions:
                    for condition in job.status.conditions:
                        if (
                            condition.type == "Complete"
                            and condition.status == "True"
                        ):
                            node_status = NodeStatus.COMPLETED
                            break
                        elif (
                            condition.type == "Failed"
                            and condition.status == "True"
                        ):
                            node_status = NodeStatus.FAILED
                            break

                node.status = node_status
                logger.debug(
                    "Existing job for step `%s` status: %s.",
                    step_name_annotation,
                    node_status,
                )

    return list(nodes.values())


def main() -> None:
    """Entrypoint of the k8s master/orchestrator pod.

    Raises:
        RuntimeError: If the orchestrator pod is not associated with a job.
    """
    args = parse_args()

    orchestrator_pod_name = socket.gethostname()

    client = Client()
    snapshot = client.get_snapshot(args.snapshot_id)

    logs_context: ContextManager[Any] = nullcontext()
    if is_pipeline_logging_enabled(snapshot.pipeline_configuration):
        logs_context = setup_logging_context(source="orchestrator")

    with logs_context:
        logger.info("Orchestrator pod started.")

        active_stack = client.active_stack
        orchestrator = active_stack.orchestrator
        assert isinstance(orchestrator, KubernetesOrchestrator)
        namespace = orchestrator.config.kubernetes_namespace

        pipeline_settings = cast(
            KubernetesOrchestratorSettings,
            orchestrator.get_settings(snapshot),
        )

        # Get a Kubernetes client from the active Kubernetes orchestrator, but
        # override the `incluster` setting to `True` since we are running inside
        # the Kubernetes cluster.
        api_client_config = orchestrator.get_kube_client(
            incluster=True
        ).configuration
        api_client_config.connection_pool_maxsize = (
            pipeline_settings.max_parallelism
        )
        kube_client = k8s_client.ApiClient(api_client_config)
        core_api = k8s_client.CoreV1Api(kube_client)
        batch_api = k8s_client.BatchV1Api(kube_client)

        job_name = kube_utils.get_parent_job_name(
            core_api=core_api,
            pod_name=orchestrator_pod_name,
            namespace=namespace,
            api_request_timeout=pipeline_settings.api_request_timeout,
        )
        if not job_name:
            raise RuntimeError(
                "Failed to fetch job name for orchestrator pod."
            )

        run_id, orchestrator_run_id = _get_orchestrator_job_state(
            batch_api=batch_api,
            namespace=namespace,
            job_name=job_name,
            api_request_timeout=pipeline_settings.api_request_timeout,
        )

        if run_id and orchestrator_run_id:
            logger.info("Continuing existing run `%s`.", run_id)
            pipeline_run = client.get_pipeline_run(run_id)
            nodes = _reconstruct_nodes(
                snapshot=snapshot,
                pipeline_run=pipeline_run,
                namespace=namespace,
                batch_api=batch_api,
                api_request_timeout=pipeline_settings.api_request_timeout,
            )
            logger.debug("Reconstructed nodes: %s", nodes)

        else:
            orchestrator_run_id = orchestrator_pod_name

            if args.run_id:
                pipeline_run = client.zen_store.update_run(
                    run_id=args.run_id,
                    run_update=PipelineRunUpdate(
                        orchestrator_run_id=orchestrator_run_id
                    ),
                )
            else:
                pipeline_run = create_placeholder_run(
                    snapshot=snapshot,
                    orchestrator_run_id=orchestrator_run_id,
                )

            # Store in the job annotations so we can continue the run if the pod
            # is restarted
            kube_utils.update_job(
                batch_api=batch_api,
                namespace=namespace,
                job_name=job_name,
                annotations={
                    RUN_ID_ANNOTATION_KEY: str(pipeline_run.id),
                    ORCHESTRATOR_RUN_ID_ANNOTATION_KEY: orchestrator_run_id,
                },
                api_request_timeout=pipeline_settings.api_request_timeout,
            )
            nodes = [
                Node(id=step_name, upstream_nodes=step.spec.upstream_steps)
                for step_name, step in snapshot.step_configurations.items()
            ]

        if isinstance(logs_context, LoggingContext):
            logs_context.update(pipeline_run=pipeline_run)
            logs_context.attach(pipeline_run)

        step_command = StepEntrypointConfiguration.get_entrypoint_command()
        mount_local_stores = active_stack.orchestrator.config.is_local

        shared_env, secrets = get_config_environment_vars()
        shared_env[ENV_ZENML_KUBERNETES_RUN_ID] = orchestrator_run_id

        owner_references = None
        if not orchestrator.config.skip_owner_references:
            try:
                owner_references = kube_utils.get_pod_owner_references(
                    core_api=core_api,
                    pod_name=orchestrator_pod_name,
                    namespace=namespace,
                    api_request_timeout=pipeline_settings.api_request_timeout,
                )
                # Make sure None of the owner references are marked as
                # controllers of the created pod, which messes with the
                # garbage collection logic.
                for owner_reference in owner_references:
                    owner_reference.controller = False
            except Exception as e:
                logger.warning(f"Failed to get pod owner references: {str(e)}")

        step_run_request_factory = StepRunRequestFactory(
            snapshot=snapshot,
            pipeline_run=pipeline_run,
            stack=active_stack,
        )
        step_runs: Dict[str, StepRunResponse] = {}

        base_labels = {
            "project_id": kube_utils.sanitize_label(str(snapshot.project_id)),
            "run_id": kube_utils.sanitize_label(str(pipeline_run.id)),
            "run_name": kube_utils.sanitize_label(str(pipeline_run.name)),
            "pipeline": kube_utils.sanitize_label(
                snapshot.pipeline_configuration.name
            ),
        }

        step_run_skip_hb_set = set()

        def _cache_step_run_if_possible(step_name: str) -> bool:
            if not step_run_request_factory.has_caching_enabled(step_name):
                return False

            step_run_request = step_run_request_factory.create_request(
                step_name
            )
            try:
                step_run_request_factory.populate_request(
                    step_run_request, step_runs=step_runs
                )
            except Exception as e:
                logger.error(
                    f"Failed to populate step run request for step {step_name}: {e}"
                )
                return False

            if step_run_request.status == ExecutionStatus.CACHED:
                step_run = publish_cached_step_run(
                    step_run_request, pipeline_run
                )
                step_runs[step_name] = step_run
                logger.info("Using cached version of step `%s`.", step_name)
                return True

            return False

        def _maybe_publish_failed_step_run(step_name: str) -> None:
            steps = Client().list_run_steps(
                name=step_name, pipeline_run_id=pipeline_run.id
            )
            if steps.total > 0:
                # Step run already exists, we don't need to publish a new one
                return

            step_run_request = step_run_request_factory.create_request(
                step_name
            )
            try:
                step_run_request_factory.populate_request(
                    step_run_request, step_runs=step_runs
                )
            except Exception as e:
                logger.error(
                    "Failed to populate step run request for step `%s`: %s",
                    step_name,
                    e,
                )
                return

            step_run_request.status = ExecutionStatus.FAILED
            step_run_request.end_time = utc_now()

            try:
                Client().zen_store.create_run_step(step_run_request)
            except Exception as e:
                logger.error(
                    "Failed to publish failed step run `%s`: %s", step_name, e
                )

        startup_lock = threading.Lock()
        last_startup_time: float = 0.0

        def start_step_job(node: Node) -> NodeStatus:
            """Run a pipeline step in a separate Kubernetes pod.

            Args:
                node: The node to start.

            Returns:
                The status of the node.
            """
            step_name = node.id
            step_config = snapshot.step_configurations[step_name].config
            settings = step_config.settings.get(
                "orchestrator.kubernetes", None
            )
            settings = KubernetesOrchestratorSettings.model_validate(
                settings.model_dump() if settings else {}
            )
            if not pipeline_settings.prevent_orchestrator_pod_caching:
                if _cache_step_run_if_possible(step_name):
                    return NodeStatus.COMPLETED

            step_labels = base_labels.copy()
            step_labels["step_name"] = kube_utils.sanitize_label(step_name)
            step_annotations = {
                STEP_NAME_ANNOTATION_KEY: step_name,
            }

            step_env = shared_env.copy()
            step_env.update(
                env_utils.get_runtime_environment(
                    config=step_config, stack=active_stack
                )
            )

            image = KubernetesOrchestrator.get_image(
                snapshot=snapshot, step_name=step_name
            )
            step_args = StepEntrypointConfiguration.get_entrypoint_arguments(
                step_name=step_name, snapshot_id=snapshot.id
            )

            # We set some default minimum memory resource requests for the step pod
            # here if the user has not specified any, because the step pod takes up
            # some memory resources itself and, if not specified, the pod will be
            # scheduled on any node regardless of available memory and risk
            # negatively impacting or even crashing the node due to memory pressure.
            pod_settings = kube_utils.apply_default_resource_requests(
                memory="400Mi",
                pod_settings=settings.pod_settings,
            )

            if orchestrator.config.pass_zenml_token_as_secret:
                step_env.pop("ZENML_STORE_API_TOKEN", None)
                secret_name = orchestrator.get_token_secret_name(snapshot.id)
                pod_settings.env.append(
                    {
                        "name": "ZENML_STORE_API_TOKEN",
                        "valueFrom": {
                            "secretKeyRef": {
                                "name": secret_name,
                                "key": KUBERNETES_SECRET_TOKEN_KEY_NAME,
                            }
                        },
                    }
                )
            else:
                step_env.update(secrets)

            pod_manifest = build_pod_manifest(
                pod_name=None,
                image_name=image,
                command=step_command,
                args=step_args,
                env=step_env,
                privileged=settings.privileged,
                pod_settings=pod_settings,
                service_account_name=settings.step_pod_service_account_name
                or settings.service_account_name,
                mount_local_stores=mount_local_stores,
                termination_grace_period_seconds=settings.pod_stop_grace_period,
                labels=step_labels,
            )

            retry_config = step_config.retry
            backoff_limit = (
                retry_config.max_retries if retry_config else 0
            ) + settings.backoff_limit_margin

            pod_failure_policy = settings.pod_failure_policy or {
                # These rules are applied sequentially. This means any failure in
                # the main container will count towards the max retries. Any other
                # disruption will not count towards the max retries.
                "rules": [
                    # If the main container fails, we count it towards the max
                    # retries.
                    {
                        "action": "Count",
                        "onExitCodes": {
                            "containerName": "main",
                            "operator": "NotIn",
                            "values": [0],
                        },
                    },
                    # If the pod is interrupted at any other time, we don't count
                    # it as a retry
                    {
                        "action": "Ignore",
                        "onPodConditions": [
                            {
                                "type": "DisruptionTarget",
                            }
                        ],
                    },
                ]
            }

            job_name = settings.job_name_prefix or ""
            random_prefix = "".join(random.choices("0123456789abcdef", k=8))
            job_name += f"-{random_prefix}-{step_name}-{snapshot.pipeline_configuration.name}"
            # The job name will be used as a label on the pods, so we need to make
            # sure it doesn't exceed the label length limit
            job_name = kube_utils.sanitize_label(job_name)

            job_manifest = build_job_manifest(
                job_name=job_name,
                pod_template=pod_template_manifest_from_pod(pod_manifest),
                backoff_limit=backoff_limit,
                ttl_seconds_after_finished=settings.ttl_seconds_after_finished,
                active_deadline_seconds=settings.active_deadline_seconds,
                pod_failure_policy=pod_failure_policy,
                owner_references=owner_references,
                labels=step_labels,
                annotations=step_annotations,
            )

            if (
                startup_interval
                := orchestrator.config.parallel_step_startup_waiting_period
            ):
                nonlocal last_startup_time

                with startup_lock:
                    now = time.time()
                    time_since_last_startup = now - last_startup_time
                    sleep_time = startup_interval - time_since_last_startup
                    if sleep_time > 0:
                        logger.debug(
                            f"Sleeping for {sleep_time} seconds before "
                            f"starting job for step {step_name}."
                        )
                        time.sleep(sleep_time)
                    last_startup_time = now

            kube_utils.create_job(
                batch_api=batch_api,
                namespace=namespace,
                job_manifest=job_manifest,
                api_request_timeout=settings.api_request_timeout,
            )

            try:
                Client().create_run_metadata(
                    metadata={"step_jobs": {step_name: job_name}},
                    resources=[
                        RunMetadataResource(
                            id=pipeline_run.id,
                            type=MetadataResourceTypes.PIPELINE_RUN,
                        )
                    ],
                )
            except Exception as e:
                logger.warning(
                    "Failed to create run metadata for step `%s`: %s",
                    step_name,
                    str(e),
                )

            node.metadata["job_name"] = job_name

            return NodeStatus.RUNNING

        def stop_step(node: Node) -> None:
            """Stop a step.

            Args:
                node: The node to stop.
            """
            step_name = node.id

            label_selector = (
                f"run_id={kube_utils.sanitize_label(str(pipeline_run.id))},"
                f"step_name={kube_utils.sanitize_label(node.id)}"
            )

            try:
                jobs = batch_api.list_namespaced_job(
                    namespace=namespace,
                    label_selector=label_selector,
                )
            except Exception as e:
                logger.warning(
                    f"Failed to find jobs for step {step_name}: {e}"
                )
                return

            if not jobs.items:
                logger.debug(f"No jobs found for step {step_name}")
                return

            for job in jobs.items:
                # Check if this is a job related to the correct step
                step_name_annotation = job.metadata.annotations.get(
                    STEP_NAME_ANNOTATION_KEY, None
                )
                if (
                    step_name_annotation is None
                    or step_name_annotation != step_name
                ):
                    continue

                # Check if job is already completed or failed - skip deletion
                job_finished = False
                if job.status.conditions:
                    for condition in job.status.conditions:
                        if (
                            condition.type in ["Complete", "Failed"]
                            and condition.status == "True"
                        ):
                            job_finished = True
                            break

                if job_finished:
                    logger.debug(
                        f"Job {job.metadata.name} for step {step_name} already finished, skipping deletion"
                    )
                    continue

                # Delete the running/pending job
                try:
                    batch_api.delete_namespaced_job(
                        name=job.metadata.name,
                        namespace=namespace,
                        propagation_policy="Foreground",
                    )
                    logger.info(
                        f"Successfully stopped step job: {job.metadata.name}"
                    )
                    break
                except Exception as e:
                    logger.warning(
                        f"Failed to stop step job {job.metadata.name}: {e}"
                    )
                    break

        def is_node_heartbeat_unhealthy(node: Node) -> bool:
            from zenml.steps.heartbeat import is_heartbeat_unhealthy

            if node.id in step_run_skip_hb_set:
                return False

            sr_ = client.list_run_steps(
                name=node.id, pipeline_run_id=pipeline_run.id, hydrate=False
            )

            if sr_.items:
                if (
                    sr_.items[0].heartbeat_threshold is None
                ):  # heartbeat disabled/unset - skip next checks
                    step_run_skip_hb_set.add(node.id)
                    return False

                return is_heartbeat_unhealthy(
                    step_run_id=sr_.items[0].id,
                    status=sr_.items[0].status,
                    start_time=sr_.items[0].start_time,
                    heartbeat_threshold=sr_.items[0].heartbeat_threshold,
                    latest_heartbeat=sr_.items[0].latest_heartbeat,
                )

            return False

        def check_job_status(node: Node) -> NodeStatus:
            """Check the status of a job.

            Args:
                node: The node to check.

            Returns:
                The status of the node.
            """
            step_name = node.id
            job_name = node.metadata.get("job_name", None)
            if not job_name:
                logger.error(
                    "Missing job name to monitor step `%s`.", step_name
                )
                return NodeStatus.FAILED

            step_config = snapshot.step_configurations[step_name].config
            settings = step_config.settings.get(
                "orchestrator.kubernetes", None
            )
            settings = KubernetesOrchestratorSettings.model_validate(
                settings.model_dump() if settings else {}
            )
            status, error_message = kube_utils.check_job_status(
                batch_api=batch_api,
                core_api=core_api,
                namespace=namespace,
                job_name=job_name,
                fail_on_container_waiting_reasons=settings.fail_on_container_waiting_reasons,
                api_request_timeout=settings.api_request_timeout,
            )
            if status == kube_utils.JobStatus.SUCCEEDED:
                return NodeStatus.COMPLETED
            elif status == kube_utils.JobStatus.FAILED:
                logger.error(
                    "Job for step `%s` failed: %s",
                    step_name,
                    error_message,
                )
                _maybe_publish_failed_step_run(step_name)
                return NodeStatus.FAILED
            elif (
                snapshot.pipeline_configuration.enable_heartbeat
                and is_node_heartbeat_unhealthy(node)
            ):
                logger.error(
                    "Heartbeat for step `%s` indicates unhealthy status.",
                    step_name,
                )
                stop_step(node=node)
                return NodeStatus.FAILED
            else:
                return NodeStatus.RUNNING

        def should_interrupt_execution() -> Optional[InterruptMode]:
            """Check if the DAG execution should be interrupted.

            Returns:
                If the DAG execution should be interrupted.
            """
            try:
                run = client.get_pipeline_run(
                    name_id_or_prefix=pipeline_run.id,
                    project=pipeline_run.project_id,
                    hydrate=False,  # We only need status, not full hydration
                )

                if run.status in [
                    ExecutionStatus.STOPPING,
                    ExecutionStatus.STOPPED,
                ]:
                    logger.info(
                        "Stopping DAG execution because pipeline run is in "
                        "`%s` state.",
                        run.status,
                    )
                    return InterruptMode.GRACEFUL
                elif run.status == ExecutionStatus.FAILED:
                    if (
                        snapshot.pipeline_configuration.execution_mode
                        == ExecutionMode.STOP_ON_FAILURE
                    ):
                        logger.info(
                            "Stopping DAG execution because pipeline run is in "
                            "`%s` state.",
                            run.status,
                        )
                        return InterruptMode.GRACEFUL
                    elif (
                        snapshot.pipeline_configuration.execution_mode
                        == ExecutionMode.FAIL_FAST
                    ):
                        logger.info(
                            "Stopping DAG execution because pipeline run is in "
                            "`%s` state.",
                            run.status,
                        )
                        return InterruptMode.FORCE
            except Exception as e:
                logger.warning(
                    "Failed to check pipeline cancellation status: %s", e
                )

            return None

        try:
            nodes_statuses = DagRunner(
                nodes=nodes,
                node_startup_function=start_step_job,
                node_monitoring_function=check_job_status,
                interrupt_function=should_interrupt_execution,
                monitoring_interval=pipeline_settings.job_monitoring_interval,
                monitoring_delay=pipeline_settings.job_monitoring_delay,
                interrupt_check_interval=pipeline_settings.interrupt_check_interval,
                max_parallelism=pipeline_settings.max_parallelism,
                node_stop_function=stop_step,
            ).run()
        finally:
            if (
                orchestrator.config.pass_zenml_token_as_secret
                and snapshot.schedule is None
            ):
                secret_name = orchestrator.get_token_secret_name(snapshot.id)
                try:
                    kube_utils.delete_secret(
                        core_api=core_api,
                        namespace=namespace,
                        secret_name=secret_name,
                    )
                except ApiException as e:
                    logger.error(
                        f"Error cleaning up secret {secret_name}: {e}"
                    )

        try:
            pipeline_failed = False
            failed_step_names = [
                step_name
                for step_name, node_state in nodes_statuses.items()
                if node_state == NodeStatus.FAILED
            ]
            skipped_step_names = [
                step_name
                for step_name, node_state in nodes_statuses.items()
                if node_state == NodeStatus.SKIPPED
            ]

            if failed_step_names:
                logger.error(
                    "The following steps failed: %s",
                    ", ".join(failed_step_names),
                )
            if skipped_step_names:
                logger.error(
                    "The following steps were skipped because some of their "
                    "upstream steps failed: %s",
                    ", ".join(skipped_step_names),
                )

            step_runs = fetch_step_runs_by_names(
                step_run_names=failed_step_names, pipeline_run=pipeline_run
            )

            for step_name, node_state in nodes_statuses.items():
                if node_state != NodeStatus.FAILED:
                    continue

                pipeline_failed = True

                if step_run := step_runs.get(step_name, None):
                    # Try to update the step run status, if it exists and is in
                    # a transient state.
                    if step_run and step_run.status in {
                        ExecutionStatus.INITIALIZING,
                        ExecutionStatus.RUNNING,
                    }:
                        publish_utils.publish_failed_step_run(step_run.id)

            # If any steps failed and the pipeline run is still in a transient
            # state, we need to mark it as failed.
            if pipeline_failed and pipeline_run.status in {
                ExecutionStatus.INITIALIZING,
                ExecutionStatus.RUNNING,
            }:
                publish_utils.publish_failed_pipeline_run(pipeline_run.id)
        except AuthorizationException:
            # If a step of the pipeline failed or all of them completed
            # successfully, the pipeline run will be finished and the API token
            # will be invalidated. We catch this exception and do nothing here,
            # as the pipeline run status will already have been published.
            pass

        logger.info("Orchestrator pod finished.")


if __name__ == "__main__":
    main()
